import logging
from typing import Union

import numpy as np
import pytest
import torch

from ludwig.constants import SEQUENCE
from ludwig.encoders.registry import get_encoder_cls
from tests.integration_tests.utils import ENCODERS

logger = logging.getLogger(__name__)

TEST_VOCAB_SIZE = 132
TEST_HIDDEN_SIZE = 32
TEST_STATE_SIZE = 16
TEST_EMBEDDING_SIZE = 64
TEST_NUM_FILTERS = 24
BATCH_SIZE = 2
SEQ_SIZE = 10
PARALLEL_CNN_LAYERS = 4

# encoder parameters combinations tested
encoder_parameters = {
    "vocab": [str(i) for i in range(TEST_VOCAB_SIZE)],
    "embedding_size": TEST_EMBEDDING_SIZE,
    "hidden_size": TEST_HIDDEN_SIZE,
    "num_filters": TEST_NUM_FILTERS,
    "num_layers": 1,
    "max_sequence_length": SEQ_SIZE,
    "state_size": TEST_STATE_SIZE,
    "cell_type": "rnn",
    "should_embed": True,
    "dropout": 0.0,
    "norm": None,
    "reduce_output": None,
}


@pytest.fixture(scope="module")
def input_sequence() -> torch.Tensor:
    # generates a realistic looking synthetic sequence tensor, i.e.
    # each sequence will have non-zero tokens at the beginning with
    # trailing zero tokens, including a max length token with a single
    # zero token at the end.  Example:
    # [
    #   [3, 5, 6, 0, 0, 0],
    #   [10, 11, 12, 13, 14, 0],   # max length sequence
    #   [32, 0, 0, 0, 0, 0]        # minimum length sequence
    # ]
    input_tensor = torch.zeros([BATCH_SIZE, SEQ_SIZE], dtype=torch.int32)
    sequence_lengths = np.random.randint(1, SEQ_SIZE, size=BATCH_SIZE)
    for i in range(input_tensor.shape[0]):
        input_tensor[i, : sequence_lengths[i]] = torch.tensor(
            np.random.randint(2, TEST_VOCAB_SIZE, size=sequence_lengths[i])
        )

    if torch.cuda.is_available():
        input_tensor = input_tensor.cuda()

    return input_tensor


@pytest.mark.parametrize("enc_reduce_output", [None, "sum"])
@pytest.mark.parametrize("enc_norm", [None, "batch", "layer"])
@pytest.mark.parametrize("enc_num_layers", [1, 2])
@pytest.mark.parametrize("enc_dropout", [0, 0.2])
@pytest.mark.parametrize("enc_cell_type", ["rnn", "gru", "lstm"])
@pytest.mark.parametrize("enc_encoder", ENCODERS + ["passthrough"])
def test_sequence_encoders(
    enc_encoder: str,
    enc_cell_type: str,
    enc_dropout: float,
    enc_num_layers: int,
    enc_norm: Union[None, str],
    enc_reduce_output: Union[None, str],
    input_sequence: torch.Tensor,
):
    # update encoder parameters for specific unit test case
    encoder_parameters["cell_type"] = enc_cell_type
    encoder_parameters["dropout"] = enc_dropout
    encoder_parameters["num_layers"] = enc_num_layers
    encoder_parameters["norm"] = enc_norm
    encoder_parameters["reduce_output"] = enc_reduce_output

    # retrieve encoder to test
    encoder_obj = get_encoder_cls(SEQUENCE, enc_encoder)(**encoder_parameters)
    if torch.cuda.is_available():
        encoder_obj = encoder_obj.cuda()

    encoder_out = encoder_obj(input_sequence)

    assert "encoder_output" in encoder_out
    assert isinstance(encoder_out["encoder_output"], torch.Tensor)

    if enc_encoder == "parallel_cnn":
        number_parallel_cnn_layers = PARALLEL_CNN_LAYERS
        output_dimension = encoder_parameters["num_filters"] * number_parallel_cnn_layers
        assert (
            encoder_out["encoder_output"].shape == (BATCH_SIZE, SEQ_SIZE, output_dimension)
            if enc_reduce_output is None
            else (BATCH_SIZE, output_dimension)
        )

    elif enc_encoder == "stacked_parallel_cnn":
        number_parallel_cnn_layers = PARALLEL_CNN_LAYERS
        output_dimension = encoder_parameters["num_filters"] * number_parallel_cnn_layers
        assert (
            encoder_out["encoder_output"].shape == (BATCH_SIZE, SEQ_SIZE, output_dimension)
            if enc_reduce_output is None
            else (BATCH_SIZE, output_dimension)
        )

    elif enc_encoder == "rnn":
        assert (
            encoder_out["encoder_output"].shape == (BATCH_SIZE, SEQ_SIZE, TEST_STATE_SIZE)
            if enc_reduce_output is None
            else (BATCH_SIZE, TEST_STATE_SIZE)
        )

        assert "encoder_output_state" in encoder_out
        if enc_cell_type == "lstm":
            assert isinstance(encoder_out["encoder_output_state"], tuple)
            assert isinstance(encoder_out["encoder_output_state"][0], torch.Tensor)
            assert isinstance(encoder_out["encoder_output_state"][1], torch.Tensor)
            assert encoder_out["encoder_output_state"][0].shape == (BATCH_SIZE, TEST_STATE_SIZE)
            assert encoder_out["encoder_output_state"][1].shape == (BATCH_SIZE, TEST_STATE_SIZE)
        else:
            assert isinstance(encoder_out["encoder_output_state"], torch.Tensor)
            assert encoder_out["encoder_output_state"].shape == (BATCH_SIZE, TEST_STATE_SIZE)

    elif enc_encoder == "cnnrnn":
        assert encoder_out["encoder_output"].shape[1:] == encoder_obj.output_shape
        assert "encoder_output_state" in encoder_out

        if enc_cell_type == "lstm":
            assert isinstance(encoder_out["encoder_output_state"], tuple)
            assert encoder_out["encoder_output_state"][0].shape == (BATCH_SIZE, TEST_STATE_SIZE)
            assert encoder_out["encoder_output_state"][1].shape == (BATCH_SIZE, TEST_STATE_SIZE)
        else:
            assert isinstance(encoder_out["encoder_output_state"], torch.Tensor)
            assert encoder_out["encoder_output_state"].shape == (BATCH_SIZE, TEST_STATE_SIZE)

    elif enc_encoder == "stacked_cnn":
        assert encoder_out["encoder_output"].shape[1:] == encoder_obj.output_shape

    elif enc_encoder == "embed":
        assert (
            encoder_out["encoder_output"].shape == (BATCH_SIZE, SEQ_SIZE, TEST_EMBEDDING_SIZE)
            if enc_reduce_output is None
            else (BATCH_SIZE, TEST_EMBEDDING_SIZE)
        )

    elif enc_encoder == "transformer":
        assert encoder_out["encoder_output"].shape[1:] == encoder_obj.output_shape

    elif enc_encoder == "passthrough":
        assert (
            encoder_out["encoder_output"].shape == (BATCH_SIZE, SEQ_SIZE, 1)
            if enc_reduce_output is None
            else (BATCH_SIZE, 1)
        )

    else:
        raise ValueError(f"{enc_encoder} is an invalid encoder specification")


@pytest.mark.parametrize("enc_reduce_output", [None, "sum", "last", "mean", "max", "concat"])
def test_passthrough_encoder(enc_reduce_output, input_sequence):
    encoder_parameters = {"reduce_output": enc_reduce_output}

    # retrieve encoder to test
    encoder_obj = get_encoder_cls(SEQUENCE, "passthrough")(**encoder_parameters)

    encoder_out = encoder_obj(input_sequence)

    assert "encoder_output" in encoder_out
    assert (
        encoder_out["encoder_output"].shape == (BATCH_SIZE, SEQ_SIZE, 1)
        if enc_reduce_output is None
        else (BATCH_SIZE, 1)
    )


# test to ensure correct handling of vocab_size and embedding_size specifications
@pytest.mark.parametrize("enc_embedding_size", [TEST_VOCAB_SIZE - 8, TEST_VOCAB_SIZE, TEST_VOCAB_SIZE + 8])
def test_sequence_embed_encoder(enc_embedding_size: int, input_sequence: torch.Tensor) -> None:
    encoder_parameters["embedding_size"] = enc_embedding_size

    encoder_obj = get_encoder_cls(SEQUENCE, "embed")(**encoder_parameters)

    encoder_out = encoder_obj(input_sequence)

    assert encoder_out["encoder_output"].size()[1:] == encoder_obj.output_shape



import pytest

try:
    from ludwig.automl.auto_tune_config import reduce_text_feature_max_length
except ImportError:
    pass


@pytest.mark.distributed
def test_reduce_text_model_mem_99ptile():
    config = {"input_features": [{"name": "description", "column": "description", "type": "text", "encoder": "bert"}]}
    training_set_metadata = {"description": {"max_sequence_length_99ptile": 117.0}}
    config_upd = {
        "input_features": [{"name": "description", "column": "description", "type": "text", "encoder": "bert"}],
        "preprocessing": {"text": {"max_sequence_length": 117}},
    }
    reduce_text_feature_max_length(config, training_set_metadata)
    assert config == config_upd


@pytest.mark.distributed
def test_reduce_text_model_mem_128():
    config = {"input_features": [{"name": "description", "column": "description", "type": "text", "encoder": "bert"}]}
    training_set_metadata = {"description": {"max_sequence_length_99ptile": 512.0}}
    config_upd = {
        "input_features": [{"name": "description", "column": "description", "type": "text", "encoder": "bert"}],
        "preprocessing": {"text": {"max_sequence_length": 128}},
    }
    reduce_text_feature_max_length(config, training_set_metadata)
    assert config == config_upd


@pytest.mark.distributed
def test_reduce_text_model_mem_override():
    config = {
        "input_features": [{"name": "description", "column": "description", "type": "text", "encoder": "bert"}],
        "preprocessing": {"text": {"max_sequence_length": 256}},
    }
    training_set_metadata = {"description": {"max_sequence_length_99ptile": 117.0}}
    config_upd = {
        "input_features": [{"name": "description", "column": "description", "type": "text", "encoder": "bert"}],
        "preprocessing": {"text": {"max_sequence_length": 117}},
    }
    reduce_text_feature_max_length(config, training_set_metadata)
    assert config == config_upd


@pytest.mark.distributed
def test_reduce_text_model_mem_respect():
    config = {
        "input_features": [{"name": "description", "column": "description", "type": "text", "encoder": "bert"}],
        "preprocessing": {"text": {"max_sequence_length": 56}},
    }
    training_set_metadata = {"description": {"max_sequence_length_99ptile": 117.0}}
    config_upd = {
        "input_features": [{"name": "description", "column": "description", "type": "text", "encoder": "bert"}],
        "preprocessing": {"text": {"max_sequence_length": 56}},
    }
    reduce_text_feature_max_length(config, training_set_metadata)
    assert config == config_upd



import ludwig.datasets
from ludwig.datasets.dataset_config import DatasetConfig
from ludwig.datasets.loaders.dataset_loader import DatasetLoader


def test_get_config_and_load(tmpdir):
    yosemite_config = ludwig.datasets._get_dataset_config("yosemite")
    assert isinstance(yosemite_config, DatasetConfig)

    yosemite_dataset = ludwig.datasets.get_dataset("yosemite", cache_dir=tmpdir)
    assert isinstance(yosemite_dataset, DatasetLoader)
    df = yosemite_dataset.load()
    assert df is not None
    assert len(df) == 18721  # Expected number of rows in Yosemite temperatures dataset.


def test_get_config_kaggle(tmpdir):
    twitter_bots_config = ludwig.datasets._get_dataset_config("twitter_bots")
    assert isinstance(twitter_bots_config, DatasetConfig)

    twitter_bots_dataset = ludwig.datasets.get_dataset("twitter_bots", cache_dir=tmpdir)
    # Twitter bots dataset is large, so we won't load it in this unit test.
    assert isinstance(twitter_bots_dataset, DatasetLoader)
    assert twitter_bots_dataset.is_kaggle_dataset



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import gzip
import logging
import os
import shutil
import tarfile
from enum import Enum
from typing import List, Optional
from zipfile import ZipFile

from ludwig.utils.fs_utils import upload_output_directory

logger = logging.getLogger(__name__)


class ArchiveType(str, Enum):
    """The type of file archive."""

    UNKNOWN = "unknown"
    ZIP = "zip"
    GZIP = "gz"
    TAR = "tar"
    TAR_ZIP = "tar.z"
    TAR_BZ2 = "tar.bz2"
    TAR_GZ = "tar.gz"


def infer_archive_type(archive_path):
    """Try to infer archive type from file extension."""
    # Get the path extension including multiple extensions, ex. ".tar.gz"
    extension = ".".join(["", *os.path.basename(archive_path).split(".")[1:]])
    extension = extension.lower()
    if extension.endswith(".tar.z") or extension.endswith(".tar.zip"):
        return ArchiveType.TAR_ZIP
    elif extension.endswith(".tar.bz2") or extension.endswith(".tbz2"):
        return ArchiveType.TAR_BZ2
    elif extension.endswith(".tar.gz") or extension.endswith(".tgz"):
        return ArchiveType.TAR_GZ
    elif extension.endswith(".tar"):
        return ArchiveType.TAR
    elif extension.endswith(".zip") or extension.endswith(".zipx"):
        return ArchiveType.ZIP
    elif extension.endswith(".gz") or extension.endswith(".gzip"):
        return ArchiveType.GZIP
    else:
        return ArchiveType.UNKNOWN


def is_archive(path):
    """Does this path a supported archive type."""
    return infer_archive_type(path) != ArchiveType.UNKNOWN


def list_archive(archive_path, archive_type: Optional[ArchiveType] = None) -> List[str]:
    """Return list of files extracted in an archive (without extracting them)."""
    if archive_type is None:
        archive_type = infer_archive_type(archive_path)
    if archive_type == ArchiveType.UNKNOWN:
        logger.error(
            f"Could not infer type of archive {archive_path}.  May be an unsupported archive type."
            "Specify archive_type in the dataset config if this file has an unknown file extension."
        )
        return []
    if archive_type == ArchiveType.ZIP:
        with ZipFile(archive_path) as zfile:
            return zfile.namelist()
    elif archive_type == ArchiveType.GZIP:
        return [".".join(archive_path.split(".")[:-1])]  # Path minus the .gz extension
    elif archive_type in {ArchiveType.TAR, ArchiveType.TAR_ZIP, ArchiveType.TAR_BZ2, ArchiveType.TAR_GZ}:
        with tarfile.open(archive_path) as tar_file:
            return tar_file.getnames()
    else:
        logger.error(f"Unsupported archive: {archive_path}")
    return []


def extract_archive(archive_path: str, archive_type: Optional[ArchiveType] = None) -> List[str]:
    """Extracts files from archive (into the same directory), returns a list of extracted files.

    Args:
        archive_path - The full path to the archive.

    Returns A list of the files extracted.
    """
    if archive_type is None:
        archive_type = infer_archive_type(archive_path)
    if archive_type == ArchiveType.UNKNOWN:
        logger.error(
            f"Could not infer type of archive {archive_path}.  May be an unsupported archive type."
            "Specify archive_type in the dataset config if this file has an unknown file extension."
        )
        return []
    archive_directory = os.path.dirname(archive_path)
    directory_contents_before = os.listdir(archive_directory)
    with upload_output_directory(archive_directory) as (tmpdir, _):
        if archive_type == ArchiveType.ZIP:
            with ZipFile(archive_path) as zfile:
                zfile.extractall(tmpdir)
        elif archive_type == ArchiveType.GZIP:
            gzip_content_file = ".".join(archive_path.split(".")[:-1])  # Path minus the .gz extension
            with gzip.open(archive_path) as gzfile:
                with open(os.path.join(tmpdir, gzip_content_file), "wb") as output:
                    shutil.copyfileobj(gzfile, output)
        elif archive_type in {ArchiveType.TAR, ArchiveType.TAR_ZIP, ArchiveType.TAR_BZ2, ArchiveType.TAR_GZ}:
            with tarfile.open(archive_path) as tar_file:

                def is_within_directory(directory, target):
                    abs_directory = os.path.abspath(directory)
                    abs_target = os.path.abspath(target)

                    prefix = os.path.commonprefix([abs_directory, abs_target])

                    return prefix == abs_directory

                def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
                    for member in tar.getmembers():
                        member_path = os.path.join(path, member.name)
                        if not is_within_directory(path, member_path):
                            raise Exception("Attempted Path Traversal in Tar File")

                    tar.extractall(path, members, numeric_owner=numeric_owner)

                safe_extract(tar_file, path=tmpdir)
        else:
            logger.error(f"Unsupported archive: {archive_path}")
    directory_contents_after = set(os.listdir(archive_directory))
    return directory_contents_after.difference(directory_contents_before)



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import RANDOM
from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import PREPROCESSING_METADATA
from ludwig.schema.split import BaseSplitConfig, SplitDataclassField
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class PreprocessingConfig(schema_utils.BaseMarshmallowConfig):
    """Global preprocessing config is a dataclass that configures the parameters used for global preprocessing."""

    sample_ratio: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="The ratio of the dataset to use. For instance, if 0.5, half of the dataset "
        "provided will be used.",
        parameter_metadata=PREPROCESSING_METADATA["sample_ratio"],
    )

    oversample_minority: float = schema_utils.NonNegativeFloat(
        default=None,
        allow_none=True,
        description="If not None, the minority class will be oversampled to reach the specified ratio respective to "
        "the majority class. ",
        parameter_metadata=PREPROCESSING_METADATA["oversample_minority"],
    )

    undersample_majority: float = schema_utils.NonNegativeFloat(
        default=None,
        allow_none=True,
        description="If not None, the majority class will be undersampled to reach the specified ratio respective "
        "to the minority class. ",
        parameter_metadata=PREPROCESSING_METADATA["undersample_majority"],
    )

    split: BaseSplitConfig = SplitDataclassField(
        default=RANDOM,
    )


@DeveloperAPI
def get_preprocessing_jsonschema():
    """Returns a JSON schema structured to only require a `type` key and then conditionally apply a corresponding
    combiner's field constraints."""
    preproc_schema = schema_utils.unload_jsonschema_from_marshmallow_class(PreprocessingConfig)
    props = preproc_schema["properties"]
    return {
        "type": "object",
        "properties": props,
        "additionalProperties": True,
        "title": "global_preprocessing_options",
        "description": "Select the preprocessing type.",
    }


@DeveloperAPI
class PreprocessingField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(PreprocessingConfig)

    @staticmethod
    def _jsonschema_type_mapping():
        return get_preprocessing_jsonschema()



import logging
import os

import numpy as np
import pandas as pd
import pytest

from ludwig.api import LudwigModel
from ludwig.constants import BATCH_SIZE, BINARY, CATEGORY, MODEL_ECD, MODEL_GBM
from ludwig.explain.captum import IntegratedGradientsExplainer
from ludwig.explain.explainer import Explainer
from ludwig.explain.explanation import Explanation
from ludwig.explain.gbm import GBMExplainer
from tests.integration_tests.utils import (
    binary_feature,
    category_feature,
    generate_data,
    image_feature,
    LocalTestBackend,
    number_feature,
    set_feature,
    text_feature,
    timeseries_feature,
    vector_feature,
)

try:
    from ludwig.explain.captum_ray import RayIntegratedGradientsExplainer
except ImportError:
    RayIntegratedGradientsExplainer = None


def test_explanation_dataclass():
    explanation = Explanation(target="target")

    feature_attributions_for_label_1 = np.array([1, 2, 3])
    feature_attributions_for_label_2 = np.array([4, 5, 6])

    # test add()
    explanation.add(["f1", "f2", "f3"], feature_attributions_for_label_1)

    with pytest.raises(AssertionError, match="Expected feature attributions of shape"):
        # test add() with wrong shape
        explanation.add(["f1", "f2", "f3", "f4"], np.array([1, 2, 3, 4]))

    explanation.add(["f1", "f2", "f3"], feature_attributions_for_label_2)

    # test to_array()
    explanation_array = explanation.to_array()
    assert np.array_equal(explanation_array, [[1, 2, 3], [4, 5, 6]])


def test_abstract_explainer_instantiation():
    with pytest.raises(TypeError, match="Can't instantiate abstract class Explainer with abstract method"):
        Explainer(None, inputs_df=None, sample_df=None, target=None)


@pytest.mark.parametrize(
    "explainer_class, model_type",
    [
        (IntegratedGradientsExplainer, MODEL_ECD),
        (GBMExplainer, MODEL_GBM),
    ],
)
@pytest.mark.parametrize(
    "output_feature",
    [binary_feature(), number_feature(), category_feature(decoder={"vocab_size": 3})],
    ids=["binary", "number", "category"],
)
@pytest.mark.parametrize(
    "additional_config",
    [
        pytest.param({}, id="default"),
        pytest.param({"preprocessing": {"split": {"type": "fixed", "column": "split"}}}, id="fixed_split"),
    ],
)
def test_explainer_api(explainer_class, model_type, output_feature, additional_config, tmpdir):
    run_test_explainer_api(explainer_class, model_type, [output_feature], additional_config, tmpdir)


@pytest.mark.distributed
@pytest.mark.parametrize(
    "output_feature",
    [binary_feature(), number_feature(), category_feature(decoder={"vocab_size": 3})],
    ids=["binary", "number", "category"],
)
def test_explainer_api_ray(output_feature, tmpdir, ray_cluster_2cpu):
    from ludwig.explain.captum_ray import RayIntegratedGradientsExplainer

    run_test_explainer_api(
        RayIntegratedGradientsExplainer,
        "ecd",
        [output_feature],
        {},
        tmpdir,
        resources_per_task={"num_cpus": 1},
        num_workers=1,
    )


@pytest.mark.parametrize("cache_encoder_embeddings", [True, False])
@pytest.mark.parametrize(
    "explainer_class,model_type",
    [
        pytest.param(IntegratedGradientsExplainer, MODEL_ECD, id="ecd_local"),
        pytest.param(RayIntegratedGradientsExplainer, MODEL_ECD, id="ecd_ray", marks=pytest.mark.distributed),
        # TODO(travis): once we support GBM text features
        # pytest.param((GBMExplainer, MODEL_GBM), id="gbm_local"),
    ],
)
def test_explainer_text_hf(explainer_class, model_type, cache_encoder_embeddings, tmpdir, ray_cluster_2cpu):
    input_features = [
        text_feature(
            encoder={
                "type": "auto_transformer",
                "pretrained_model_name_or_path": "hf-internal-testing/tiny-bert-for-token-classification",
            },
            preprocessing={"cache_encoder_embeddings": cache_encoder_embeddings},
        )
    ]
    run_test_explainer_api(explainer_class, model_type, [binary_feature()], {}, tmpdir, input_features=input_features)


def run_test_explainer_api(
    explainer_class, model_type, output_features, additional_config, tmpdir, input_features=None, **kwargs
):
    image_dest_folder = os.path.join(tmpdir, "generated_images")

    if input_features is None:
        input_features = [binary_feature(), number_feature(), category_feature(encoder={"reduce_output": "sum"})]
        if model_type == MODEL_ECD:
            input_features += [
                text_feature(encoder={"vocab_size": 3}),
                vector_feature(),
                timeseries_feature(),
                image_feature(folder=image_dest_folder),
                # audio_feature(os.path.join(tmpdir, "generated_audio")), # NOTE: works but takes a long time
                # sequence_feature(encoder={"vocab_size": 3}),
                # date_feature(),
                # h3_feature(),
                set_feature(encoder={"vocab_size": 3}),
                # bag_feature(encoder={"vocab_size": 3}),
            ]

    # Generate data
    csv_filename = os.path.join(tmpdir, "training.csv")
    generate_data(input_features, output_features, csv_filename, num_examples=200)
    df = pd.read_csv(csv_filename)
    if "split" in additional_config.get("preprocessing", {}):
        df["split"] = np.random.randint(0, 3, df.shape[0])

    # Train model
    config = {"input_features": input_features, "output_features": output_features, "model_type": model_type}
    if model_type == MODEL_ECD:
        config["trainer"] = {"epochs": 2, BATCH_SIZE: 128}
    else:
        # Disable feature filtering to avoid having no features due to small test dataset,
        # see https://stackoverflow.com/a/66405983/5222402
        config["trainer"] = {"feature_pre_filter": False}
    config.update(additional_config)

    model = LudwigModel(config, logging_level=logging.WARNING, backend=LocalTestBackend())
    model.train(df)

    # Explain model
    explainer = explainer_class(model, inputs_df=df, sample_df=df, target=output_features[0]["name"], **kwargs)

    is_binary = output_features[0].get("type") == BINARY
    is_category = output_features[0].get("type") == CATEGORY

    vocab_size = 1
    if is_binary:
        vocab_size = 2
    elif is_category:
        vocab_size = output_features[0].get("decoder", {}).get("vocab_size")

    assert explainer.is_binary_target == is_binary
    assert explainer.is_category_target == is_category
    assert explainer.vocab_size == vocab_size

    explanations_result = explainer.explain()

    # Verify shapes.
    assert explanations_result.global_explanation.to_array().shape == (vocab_size, len(input_features))

    assert len(explanations_result.row_explanations) == len(df)
    for e in explanations_result.row_explanations:
        assert e.to_array().shape == (vocab_size, len(input_features))

    assert len(explanations_result.expected_values) == vocab_size



from typing import List, Optional

import numpy as np
import pytest
import torch

from ludwig.modules.fully_connected_modules import FCLayer, FCStack
from ludwig.utils.misc_utils import set_random_seed
from ludwig.utils.torch_utils import get_torch_device

BATCH_SIZE = 2
DEVICE = get_torch_device()
RANDOM_SEED = 1919


@pytest.mark.parametrize("input_size", [2, 3])
@pytest.mark.parametrize("output_size", [3, 4])
@pytest.mark.parametrize("activation", ["relu", "sigmoid", "tanh"])
@pytest.mark.parametrize("dropout", [0.0, 0.6])
def test_fc_layer(
    input_size: int,
    output_size: int,
    activation: str,
    dropout: float,
):
    set_random_seed(RANDOM_SEED)  # make repeatable
    fc_layer = FCLayer(input_size=input_size, output_size=output_size, activation=activation, dropout=dropout).to(
        DEVICE
    )
    input_tensor = torch.randn(BATCH_SIZE, input_size, device=DEVICE)
    output_tensor = fc_layer(input_tensor)
    assert output_tensor.shape[1:] == fc_layer.output_shape


@pytest.mark.parametrize(
    "first_layer_input_size,layers,num_layers",
    [
        (2, None, 3),
        (2, [{"output_size": 4}, {"output_size": 8}], None),
        (2, [{"input_size": 2, "output_size": 4}, {"output_size": 8}], None),
    ],
)
def test_fc_stack(
    first_layer_input_size: Optional[int],
    layers: Optional[List],
    num_layers: Optional[int],
):
    set_random_seed(RANDOM_SEED)
    fc_stack = FCStack(first_layer_input_size=first_layer_input_size, layers=layers, num_layers=num_layers).to(DEVICE)
    input_tensor = torch.randn(BATCH_SIZE, first_layer_input_size, device=DEVICE)
    output_tensor = fc_stack(input_tensor)
    assert output_tensor.shape[1:] == fc_stack.output_shape


def test_fc_stack_input_size_mismatch_fails():
    first_layer_input_size = 10
    layers = [{"input_size": 2, "output_size": 4}, {"output_size": 8}]

    fc_stack = FCStack(
        first_layer_input_size=first_layer_input_size,
        layers=layers,
    ).to(DEVICE)
    input_tensor = torch.randn(BATCH_SIZE, first_layer_input_size, device=DEVICE)

    with pytest.raises(RuntimeError):
        fc_stack(input_tensor)


def test_fc_stack_no_layers_behaves_like_passthrough():
    first_layer_input_size = 10
    layers = None
    num_layers = 0
    output_size = 15

    fc_stack = FCStack(
        first_layer_input_size=first_layer_input_size,
        layers=layers,
        num_layers=num_layers,
        default_output_size=output_size,
    ).to(DEVICE)
    input_tensor = torch.randn(BATCH_SIZE, first_layer_input_size, device=DEVICE)
    output_tensor = fc_stack(input_tensor)

    assert list(output_tensor.shape[1:]) == [first_layer_input_size]
    assert output_tensor.shape[1:] == fc_stack.output_shape
    assert np.all(np.isclose(input_tensor, output_tensor))



import numpy as np
import pytest
import torch

from ludwig.features import feature_utils


def test_ludwig_feature_dict():
    feature_dict = feature_utils.LudwigFeatureDict()

    to_module = torch.nn.Module()
    type_module = torch.nn.Module()

    feature_dict.set("to", to_module)
    feature_dict.set("type", type_module)

    assert iter(feature_dict) is not None
    assert next(feature_dict) is not None
    assert len(feature_dict) == 2
    assert feature_dict.keys() == ["to", "type"]
    assert feature_dict.items() == [("to", to_module), ("type", type_module)]
    assert feature_dict.get("to"), to_module

    feature_dict.update({"to_empty": torch.nn.Module()})

    assert len(feature_dict) == 3
    assert [key for key in feature_dict] == ["to", "type", "to_empty"]


def test_ludwig_feature_dict_with_periods():
    feature_dict = feature_utils.LudwigFeatureDict()

    to_module = torch.nn.Module()

    feature_dict.set("to.", to_module)

    assert feature_dict.keys() == ["to."]
    assert feature_dict.items() == [("to.", to_module)]
    assert feature_dict.get("to.") == to_module


@pytest.mark.parametrize("sequence_type", [list, tuple, np.array])
def test_compute_token_probabilities(sequence_type):
    inputs = sequence_type(
        [
            [0.1, 0.2, 0.7],
            [0.3, 0.4, 0.3],
            [0.6, 0.3, 0.2],
        ]
    )

    token_probabilities = feature_utils.compute_token_probabilities(inputs)
    assert np.allclose(token_probabilities, [0.7, 0.4, 0.6])


def test_compute_sequence_probability():
    inputs = np.array([0.7, 0.4, 0.6])

    sequence_probability = feature_utils.compute_sequence_probability(
        inputs, max_sequence_length=2, return_log_prob=False
    )

    assert np.allclose(sequence_probability, [0.28])  # 0.7 * 0.4



from typing import Optional

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import utils as schema_utils
from ludwig.schema.combiners.base import BaseCombinerConfig
from ludwig.schema.combiners.utils import CombinerSelection
from ludwig.schema.defaults.ecd import ECDDefaultsConfig, ECDDefaultsField
from ludwig.schema.features.base import (
    BaseInputFeatureConfig,
    BaseOutputFeatureConfig,
    ECDInputFeatureSelection,
    ECDOutputFeatureSelection,
    FeatureCollection,
)
from ludwig.schema.hyperopt import HyperoptConfig, HyperoptField
from ludwig.schema.model_types.base import ModelConfig, register_model_type
from ludwig.schema.preprocessing import PreprocessingConfig, PreprocessingField
from ludwig.schema.trainer import ECDTrainerConfig, ECDTrainerField
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@register_model_type(name="ecd")
@ludwig_dataclass
class ECDModelConfig(ModelConfig):
    """Parameters for ECD."""

    model_type: str = schema_utils.ProtectedString("ecd")

    input_features: FeatureCollection[BaseInputFeatureConfig] = ECDInputFeatureSelection().get_list_field()
    output_features: FeatureCollection[BaseOutputFeatureConfig] = ECDOutputFeatureSelection().get_list_field()

    combiner: BaseCombinerConfig = CombinerSelection().get_default_field()

    trainer: ECDTrainerConfig = ECDTrainerField().get_default_field()
    preprocessing: PreprocessingConfig = PreprocessingField().get_default_field()
    defaults: ECDDefaultsConfig = ECDDefaultsField().get_default_field()
    hyperopt: Optional[HyperoptConfig] = HyperoptField().get_default_field()



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
from typing import Optional

import pandas as pd

from ludwig.datasets.dataset_config import DatasetConfig
from ludwig.datasets.loaders.dataset_loader import DatasetLoader


class KDDCup2009Loader(DatasetLoader):
    def __init__(
        self, config: DatasetConfig, cache_dir: Optional[str] = None, task_name="", include_test_download=False
    ):
        super().__init__(config, cache_dir=cache_dir)
        self.task_name = task_name
        self.include_test_download = include_test_download

    def load_file_to_dataframe(self, file_path: str) -> pd.DataFrame:
        """Loads a file into a dataframe."""
        return pd.read_csv(file_path, sep="\t")

    def transform_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        train_df = super().transform_dataframe(dataframe)
        train_df = process_categorical_features(train_df, categorical_features)
        train_df = process_number_features(train_df, categorical_features)

        targets = (
            pd.read_csv(os.path.join(self.raw_dataset_dir, f"orange_small_train_{self.task_name}.labels"), header=None)[
                0
            ]
            .astype(str)
            .apply(lambda x: "true" if x == "1" else "false")
        )

        train_idcs = pd.read_csv(
            os.path.join(self.raw_dataset_dir, f"stratified_train_idx_{self.task_name}.txt"), header=None
        )[0]

        val_idcs = pd.read_csv(
            os.path.join(self.raw_dataset_dir, f"stratified_test_idx_{self.task_name}.txt"), header=None
        )[0]

        processed_train_df = train_df.iloc[train_idcs].copy()
        processed_train_df["target"] = targets.iloc[train_idcs]
        processed_train_df["split"] = 0

        processed_val_df = train_df.iloc[val_idcs].copy()
        processed_val_df["target"] = targets.iloc[val_idcs]
        processed_val_df["split"] = 1

        if self.include_test_download:
            test_df = self.load_file_to_dataframe(os.path.join(self.raw_dataset_dir, "orange_small_test.data"))
            test_df["target"] = ""  # no ground truth labels for test download
            test_df["split"] = 2
            df = pd.concat([processed_train_df, processed_val_df, test_df])
        else:
            df = pd.concat([processed_train_df, processed_val_df])

        return df


def process_categorical_features(df, categorical_features):
    for i in categorical_features:
        df.iloc[:, i].fillna("", inplace=True)
    return df


def process_number_features(df, categorical_features):
    for i, column in enumerate(df.columns):
        if i not in categorical_features:
            df[column].astype(float, copy=False)
    return df


categorical_features = {
    190,
    191,
    192,
    193,
    194,
    195,
    196,
    197,
    198,
    199,
    200,
    201,
    202,
    203,
    204,
    205,
    206,
    207,
    209,
    210,
    211,
    212,
    213,
    214,
    215,
    216,
    217,
    218,
    219,
    220,
    221,
    222,
    223,
    224,
    225,
    226,
    227,
    228,
}


class KDDAppetencyLoader(KDDCup2009Loader):
    """The KDD Cup 2009 Appetency dataset.

    https://www.kdd.org/kdd-cup/view/kdd-cup-2009/Data
    """

    def __init__(self, config: DatasetConfig, cache_dir: Optional[str] = None, include_test_download=False):
        super().__init__(
            config, cache_dir=cache_dir, task_name="appetency", include_test_download=include_test_download
        )


class KDDChurnLoader(KDDCup2009Loader):
    """The KDD Cup 2009 Churn dataset.

    https://www.kdd.org/kdd-cup/view/kdd-cup-2009/Data
    """

    def __init__(self, config: DatasetConfig, cache_dir: Optional[str] = None, include_test_download=False):
        super().__init__(config, cache_dir=cache_dir, task_name="churn", include_test_download=include_test_download)


class KDDUpsellingLoader(KDDCup2009Loader):
    """The KDD Cup 2009 Upselling dataset.

    https://www.kdd.org/kdd-cup/view/kdd-cup-2009/Data
    """

    def __init__(self, config: DatasetConfig, cache_dir: Optional[str] = None, include_test_download=False):
        super().__init__(
            config, cache_dir=cache_dir, task_name="upselling", include_test_download=include_test_download
        )



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import json
import logging
import os.path
from typing import Dict, List

import mlflow
import pandas as pd
import pytest
from mlflow.tracking import MlflowClient

from ludwig.backend import initialize_backend
from ludwig.callbacks import Callback
from ludwig.constants import ACCURACY, AUTO, BATCH_SIZE, EXECUTOR, MAX_CONCURRENT_TRIALS, TRAINER
from ludwig.contribs.mlflow import MlflowCallback
from ludwig.globals import HYPEROPT_STATISTICS_FILE_NAME, MODEL_HYPERPARAMETERS_FILE_NAME
from ludwig.hyperopt.results import HyperoptResults
from ludwig.hyperopt.run import hyperopt
from ludwig.hyperopt.utils import update_hyperopt_params_with_defaults
from ludwig.schema.model_config import ModelConfig
from ludwig.utils.automl.utils import get_model_type
from tests.integration_tests.utils import category_feature, generate_data, text_feature

try:
    import ray
    from ray.tune import Callback as TuneCallback
    from ray.tune.experiment.trial import Trial

    from ludwig.hyperopt.execution import get_build_hyperopt_executor
except ImportError:
    ray = None
    Trial = None
    TuneCallback = object  # needed to set up HyperoptTestCallback when not distributed


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.getLogger("ludwig").setLevel(logging.INFO)

HYPEROPT_CONFIG = {
    "parameters": {
        "trainer.learning_rate": {
            "space": "loguniform",
            "lower": 0.001,
            "upper": 0.1,
        },
        "combiner.num_fc_layers": {"space": "randint", "lower": 0, "upper": 2},
        "utterance.cell_type": {"space": "grid_search", "values": ["rnn", "gru"]},
        "utterance.fc_layers": {
            "space": "choice",
            "categories": [
                [{"output_size": 16}, {"output_size": 8}],
                [{"output_size": 16}],
                [{"output_size": 8}],
            ],
        },
    },
    "goal": "minimize",
}


SCENARIOS = [
    {"executor": {"type": "ray"}, "search_alg": {"type": "variant_generator"}},
    {"executor": {"type": "ray", "num_samples": 2}, "search_alg": {"type": "variant_generator"}},
    {
        "executor": {
            "type": "ray",
            "num_samples": 3,
            "scheduler": {
                "type": "hb_bohb",
                "time_attr": "training_iteration",
                "reduction_factor": 4,
            },
        },
        "search_alg": {"type": "bohb"},
    },
]


def _get_config(search_alg: Dict, executor: Dict, epochs: int):
    input_features = [
        text_feature(name="utterance", encoder={"cell_type": "lstm", "reduce_output": "sum"}),
        category_feature(encoder={"vocab_size": 2}, reduce_input="sum"),
    ]

    output_features = [category_feature(decoder={"vocab_size": 2}, reduce_input="sum", output_feature=True)]

    return {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat"},
        TRAINER: {"epochs": epochs, "learning_rate": 0.001, BATCH_SIZE: 128},
        "hyperopt": {
            **HYPEROPT_CONFIG,
            "executor": executor,
            "search_alg": search_alg,
        },
    }


class HyperoptTestCallback(TuneCallback):
    def __init__(self, exp_name: str, model_type: str):
        self.exp_name = exp_name
        self.model_type = model_type
        self.trial_ids = set()
        self.trial_status = {}
        self.user_config = {}
        self.rendered_config = {}

    def on_trial_start(self, iteration: int, trials: List["Trial"], trial: "Trial", **info):
        super().on_trial_start(iteration, trials, trial, **info)
        self.trial_ids.add(trial.trial_id)

    def on_trial_complete(self, iteration: int, trials: List["Trial"], trial: "Trial", **info):  # noqa
        super().on_trial_complete(iteration, trials, trial, **info)
        self.trial_status[trial.trial_id] = trial.status

        model_hyperparameters = os.path.join(
            trial.logdir, f"{self.exp_name}_{self.model_type}", "model", MODEL_HYPERPARAMETERS_FILE_NAME
        )
        if os.path.isfile(model_hyperparameters):
            try:
                with open(model_hyperparameters) as f:
                    config = json.load(f)
                    assert config, f"Trial {trial} rendered config was empty."
                self.rendered_config[trial.trial_id] = True
            except OSError:
                logging.exception("Could not load rendered config from trial logdir.")

        model_hyperparameters = os.path.join(trial.logdir, "trial_hyperparameters.json")
        if os.path.isfile(model_hyperparameters):
            try:
                with open(model_hyperparameters) as f:
                    config = json.load(f)
                    assert config, "Trial {trial} user config was empty."
                self.rendered_config[trial.trial_id] = True
            except OSError:
                logging.exception("Could not load rendered config from trial logdir.")


def run_hyperopt_executor(
    search_alg,
    executor,
    epochs,
    csv_filename,
    tmpdir,
    validate_output_feature=False,
    validation_metric=None,
    use_split=True,
):
    config = _get_config(search_alg, executor, epochs)
    rel_path = generate_data(config["input_features"], config["output_features"], csv_filename)

    if not use_split:
        df = pd.read_csv(rel_path)
        df["split"] = 0
        df.to_csv(rel_path)

    config = ModelConfig.from_dict(config).to_dict()

    hyperopt_config = config["hyperopt"]

    if validate_output_feature:
        hyperopt_config["output_feature"] = config["output_features"][0]["name"]
    if validation_metric:
        hyperopt_config["validation_metric"] = validation_metric

    backend = initialize_backend("local")
    update_hyperopt_params_with_defaults(hyperopt_config)
    if hyperopt_config[EXECUTOR].get(MAX_CONCURRENT_TRIALS) == AUTO:
        hyperopt_config[EXECUTOR][MAX_CONCURRENT_TRIALS] = backend.max_concurrent_trials(hyperopt_config)

    parameters = hyperopt_config["parameters"]
    if search_alg.get("type", "") == "bohb":
        # bohb does not support grid_search search space
        del parameters["utterance.cell_type"]
        hyperopt_config["parameters"] = parameters

    split = hyperopt_config["split"]
    output_feature = hyperopt_config["output_feature"]
    metric = hyperopt_config["metric"]
    goal = hyperopt_config["goal"]
    search_alg = hyperopt_config["search_alg"]
    executor = hyperopt_config["executor"]

    hyperopt_executor = get_build_hyperopt_executor(executor["type"])(
        parameters, output_feature, metric, goal, split, search_alg=search_alg, **executor
    )

    hyperopt_executor.execute(config, dataset=rel_path, output_directory=tmpdir, backend=backend)


@pytest.mark.distributed
@pytest.mark.parametrize("scenario", SCENARIOS)
def test_hyperopt_executor(scenario, csv_filename, tmpdir, ray_cluster_4cpu):
    search_alg = scenario["search_alg"]
    executor = scenario["executor"]
    # When using the hb_bohb scheduler, num_epochs must equal max_t (which is 81 by default)
    epochs = 2 if scenario["executor"].get("scheduler", {}).get("type", {}) != "hb_bohb" else 81
    run_hyperopt_executor(search_alg, executor, epochs, csv_filename, tmpdir)


@pytest.mark.distributed
@pytest.mark.parametrize("use_split", [True, False], ids=["split", "no_split"])
def test_hyperopt_executor_with_metric(use_split, csv_filename, tmpdir, ray_cluster_4cpu):
    run_hyperopt_executor(
        {"type": "variant_generator"},  # search_alg
        {"type": "ray", "num_samples": 2},  # executor
        2,
        csv_filename,
        tmpdir,
        validate_output_feature=True,
        validation_metric=ACCURACY,
        use_split=use_split,
    )


@pytest.mark.distributed
@pytest.mark.parametrize("backend", ["local", "ray"])
def test_hyperopt_run_hyperopt(csv_filename, backend, tmpdir, ray_cluster_4cpu):
    input_features = [
        text_feature(name="utterance", encoder={"cell_type": "lstm", "reduce_output": "sum"}),
        category_feature(encoder={"vocab_size": 2}, reduce_input="sum"),
    ]
    output_features = [category_feature(decoder={"vocab_size": 2}, reduce_input="sum", output_feature=True)]

    rel_path = generate_data(input_features, output_features, csv_filename)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat"},
        TRAINER: {"epochs": 2, "learning_rate": 0.001, BATCH_SIZE: 128},
        "backend": {
            "type": backend,
        },
    }

    output_feature_name = output_features[0]["name"]

    hyperopt_configs = {
        "parameters": {
            "trainer.learning_rate": {
                "space": "loguniform",
                "lower": 0.001,
                "upper": 0.1,
            },
            output_feature_name + ".output_size": {"space": "randint", "lower": 8, "upper": 16},
            output_feature_name + ".num_fc_layers": {"space": "randint", "lower": 0, "upper": 1},
        },
        "goal": "minimize",
        "output_feature": output_feature_name,
        "validation_metrics": "loss",
        "executor": {
            "type": "ray",
            "num_samples": 2,
            "cpu_resources_per_trial": 2,
            "max_concurrent_trials": "auto",
        },
        "search_alg": {"type": "variant_generator"},
    }

    @ray.remote(num_cpus=0)
    class Event:
        def __init__(self):
            self._set = False

        def is_set(self):
            return self._set

        def set(self):
            self._set = True

    # Used to trigger a cancel event in the trial, which should subsequently be retried
    event = Event.remote()

    class CancelCallback(Callback):
        def on_epoch_start(self, trainer, progress_tracker, save_path: str):
            if progress_tracker.epoch == 1 and not ray.get(event.is_set.remote()):
                ray.get(event.set.remote())
                raise KeyboardInterrupt()

    # add hyperopt parameter space to the config
    config["hyperopt"] = hyperopt_configs

    # run for one epoch, then cancel, then resume from where we left off
    run_hyperopt(config, rel_path, tmpdir, callbacks=[CancelCallback()])


@pytest.mark.distributed
def test_hyperopt_ray_mlflow(csv_filename, tmpdir, ray_cluster_4cpu):
    mlflow_uri = f"file://{tmpdir}/mlruns"
    mlflow.set_tracking_uri(mlflow_uri)
    client = MlflowClient(tracking_uri=mlflow_uri)

    num_samples = 2
    config = _get_config(
        {"type": "variant_generator"},  # search_alg
        {"type": "ray", "num_samples": num_samples},  # executor
        2,  # epochs
    )

    rel_path = generate_data(config["input_features"], config["output_features"], csv_filename)

    exp_name = "mlflow_test"
    run_hyperopt(config, rel_path, tmpdir, experiment_name=exp_name, callbacks=[MlflowCallback(mlflow_uri)])

    experiment = client.get_experiment_by_name(exp_name)
    assert experiment is not None

    runs = client.search_runs([experiment.experiment_id])
    assert len(runs) > 0

    for run in runs:
        artifacts = [f.path for f in client.list_artifacts(run.info.run_id, "")]
        assert "config.yaml" in artifacts
        assert "model" in artifacts


def run_hyperopt(
    config,
    rel_path,
    tmpdir,
    experiment_name="ray_hyperopt",
    callbacks=None,
):

    tune_test_callback = HyperoptTestCallback(experiment_name, get_model_type(config))

    hyperopt_results = hyperopt(
        config,
        dataset=rel_path,
        output_directory=tmpdir,
        experiment_name=experiment_name,
        callbacks=callbacks,
        tune_callbacks=[tune_test_callback],
    )

    # check for return results
    assert isinstance(hyperopt_results, HyperoptResults)

    # check for existence of the hyperopt statistics file
    assert os.path.isfile(os.path.join(tmpdir, experiment_name, HYPEROPT_STATISTICS_FILE_NAME))

    # check for evidence that the HyperoptTestCallback was active
    assert len(tune_test_callback.trial_ids) > 0
    for t in tune_test_callback.trial_ids:
        if tune_test_callback.trial_status.get(t) == "terminated":
            assert tune_test_callback.user_config[t].get()
            assert tune_test_callback.rendered_config[t].get()



import copy
import dataclasses
import json
import logging
import os
import warnings
from typing import Any, Dict

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import (
    AUTO,
    COMBINED,
    EXECUTOR,
    GOAL,
    GRID_SEARCH,
    HYPEROPT,
    INPUT_FEATURES,
    LOSS,
    MAX_CONCURRENT_TRIALS,
    METRIC,
    MINIMIZE,
    NAME,
    NUM_SAMPLES,
    OUTPUT_FEATURES,
    PARAMETERS,
    PREPROCESSING,
    RAY,
    SPACE,
    SPLIT,
    TYPE,
    VALIDATION,
)
from ludwig.globals import HYPEROPT_STATISTICS_FILE_NAME
from ludwig.hyperopt.results import HyperoptResults, TrialResults
from ludwig.types import HyperoptConfigDict, ModelConfigDict
from ludwig.utils.data_utils import save_json
from ludwig.utils.misc_utils import (
    get_class_attributes,
    get_from_registry,
    merge_dict,
    set_default_value,
    set_default_values,
)
from ludwig.utils.print_utils import print_boxed

logger = logging.getLogger(__name__)


def print_hyperopt_results(hyperopt_results: HyperoptResults):
    print_boxed("HYPEROPT RESULTS", print_fun=logger.info)
    for trial_results in hyperopt_results.ordered_trials:
        if not isinstance(trial_results.metric_score, str):
            logger.info(f"score: {trial_results.metric_score:.6f} | parameters: {trial_results.parameters}")
    logger.info("")


def save_hyperopt_stats(hyperopt_stats, hyperopt_dir_name):
    hyperopt_stats_fn = os.path.join(hyperopt_dir_name, HYPEROPT_STATISTICS_FILE_NAME)
    save_json(hyperopt_stats_fn, hyperopt_stats)


def load_json_value(v):
    try:
        return json.loads(v)
    except Exception as e:
        logger.warning(f"While loading json, encountered exception: {e}")
        return v


# define set containing names to return for TrialResults
TRIAL_RESULTS_NAMES_SET = {f.name for f in dataclasses.fields(TrialResults)}


def load_json_values(d):
    # ensure metric_score is a string for the json load to eliminate extraneous exception message
    d["metric_score"] = str(d["metric_score"])

    # load only data required for TrialResults
    return {k: load_json_value(v) for k, v in d.items() if k in TRIAL_RESULTS_NAMES_SET}


def should_tune_preprocessing(config):
    parameters = config[HYPEROPT][PARAMETERS]
    for param_name in parameters.keys():
        if f"{PREPROCESSING}." in param_name:
            return True
    return False


def parameter_to_dict(name, value):
    if name == ".":
        # Parameter name ".", means top-level config
        return value

    parameter_dict = {}
    curr_dict = parameter_dict
    name_list = name.split(".")
    for i, name_elem in enumerate(name_list):
        if i == len(name_list) - 1:
            curr_dict[name_elem] = value
        else:
            name_dict = curr_dict.get(name_elem, {})
            curr_dict[name_elem] = name_dict
            curr_dict = name_dict
    return parameter_dict


def feature_list_to_dict(config: ModelConfigDict) -> ModelConfigDict:
    input_features_dict = {}
    for feature in config[INPUT_FEATURES]:
        input_features_dict[feature[NAME]] = feature

    output_features_dict = {}
    for feature in config[OUTPUT_FEATURES]:
        output_features_dict[feature[NAME]] = feature

    config = copy.copy(config)
    config[INPUT_FEATURES] = input_features_dict
    config[OUTPUT_FEATURES] = output_features_dict
    return config


def feature_dict_to_list(config: ModelConfigDict) -> ModelConfigDict:
    # This works because Python dicts are order-preserving, so we do not need to
    # do anything special to map from a key in the dict to an index in a list
    input_features_list = []
    for feature in config[INPUT_FEATURES].values():
        input_features_list.append(feature)

    output_features_list = []
    for feature in config[OUTPUT_FEATURES].values():
        output_features_list.append(feature)

    config = copy.copy(config)
    config[INPUT_FEATURES] = input_features_list
    config[OUTPUT_FEATURES] = output_features_list
    return config


def substitute_parameters(
    config: ModelConfigDict,
    parameters: Dict[str, Any],
):
    """Update Ludwig config with parameters sampled from the Hyperopt sampler."""

    # Collect the sets of names for each feature grouping so we can map feature names to
    # groups
    input_feature_names = {feature[NAME] for feature in config[INPUT_FEATURES]}
    output_feature_names = {feature[NAME] for feature in config[OUTPUT_FEATURES]}

    # Features in the user config are provided as a list, but in hyperopt we reference
    # features by name, so convert temporarily to a dict to simplify the mergep process.
    config = feature_list_to_dict(config)

    # Merge parameters into the user configuration in order. As such, if there are conflicting
    # params, the later params will take precedence.
    for name, value in parameters.items():
        # User params are provided as <feature_name>.<param>, but we group input / output features
        # together during the merge to make it easier and unambiguous to convert back and forth
        # TODO(travis): we should revisit the user format here, as it silently breaks situations
        # where the user has a feature named "trainer", "combiner", etc.
        prefix = name.split(".")[0]
        if prefix in input_feature_names:
            name = f"{INPUT_FEATURES}.{name}"
        elif prefix in output_feature_names:
            name = f"{OUTPUT_FEATURES}.{name}"

        param_dict = parameter_to_dict(name, value)
        config = merge_dict(config, param_dict)

    # Now that all features have been merged, convert back to the original list format.
    config = feature_dict_to_list(config)

    return config


@DeveloperAPI
def get_num_duplicate_trials(hyperopt_config: HyperoptConfigDict) -> int:
    """Returns the number of duplicate trials that will be created.

    Duplicate trials are only created when there are grid type parameters and num_samples > 1.
    """
    num_samples = hyperopt_config[EXECUTOR].get(NUM_SAMPLES, 1)
    if num_samples == 1:
        return 0

    total_grid_search_trials = 1
    for _, param_info in hyperopt_config[PARAMETERS].items():
        if param_info.get(SPACE, None) == GRID_SEARCH:
            total_grid_search_trials *= len(param_info.get("values", []))

    num_duplicate_trials = (total_grid_search_trials * num_samples) - total_grid_search_trials
    return num_duplicate_trials


def log_warning_if_all_grid_type_parameters(hyperopt_config: HyperoptConfigDict) -> None:
    """Logs warning if all parameters have a grid type search space and num_samples > 1 since this will result in
    duplicate trials being created."""
    num_duplicate_trials = get_num_duplicate_trials(hyperopt_config)
    if num_duplicate_trials == 0:
        return

    num_samples = hyperopt_config[EXECUTOR].get(NUM_SAMPLES, 1)
    warnings.warn(
        "All hyperopt parameters in Ludwig config are using grid_search space, but number of samples "
        f"({num_samples}) is greater than 1. This will result in {num_duplicate_trials} duplicate trials being "
        "created. Consider setting `num_samples` to 1 in the hyperopt executor to prevent trial duplication.",
        RuntimeWarning,
    )


def update_hyperopt_params_with_defaults(hyperopt_params: HyperoptConfigDict) -> None:
    """Updates user's Ludwig config with default hyperopt parameters."""
    from ludwig.hyperopt.execution import executor_registry

    set_default_value(hyperopt_params, EXECUTOR, {})
    set_default_value(hyperopt_params, SPLIT, VALIDATION)
    set_default_value(hyperopt_params, "output_feature", COMBINED)
    set_default_value(hyperopt_params, METRIC, LOSS)
    set_default_value(hyperopt_params, GOAL, MINIMIZE)

    set_default_values(
        hyperopt_params[EXECUTOR],
        {TYPE: RAY, NUM_SAMPLES: 1, MAX_CONCURRENT_TRIALS: AUTO},
    )

    if hyperopt_params[EXECUTOR].get("trial_driver_resources") is None:
        hyperopt_params[EXECUTOR]["trial_driver_resources"] = {"CPU": 1, "GPU": 0}

    executor = get_from_registry(hyperopt_params[EXECUTOR][TYPE], executor_registry)
    executor_defaults = {k: v for k, v in executor.__dict__.items() if k in get_class_attributes(executor)}
    set_default_values(
        hyperopt_params[EXECUTOR],
        executor_defaults,
    )



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BINARY, CATEGORY, NUMBER
from ludwig.schema import utils as schema_utils
from ludwig.schema.defaults.base import BaseDefaultsConfig
from ludwig.schema.defaults.utils import DefaultsDataclassField
from ludwig.schema.features.base import BaseFeatureConfig
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class GBMDefaultsConfig(BaseDefaultsConfig):
    binary: BaseFeatureConfig = DefaultsDataclassField(feature_type=BINARY)

    category: BaseFeatureConfig = DefaultsDataclassField(feature_type=CATEGORY)

    number: BaseFeatureConfig = DefaultsDataclassField(feature_type=NUMBER)


@DeveloperAPI
def get_gbm_defaults_jsonschema():
    """Returns a JSON schema structured to only require a `type` key and then conditionally apply a corresponding
    combiner's field constraints."""
    preproc_schema = schema_utils.unload_jsonschema_from_marshmallow_class(GBMDefaultsConfig)
    props = preproc_schema["properties"]

    return {
        "type": "object",
        "properties": props,
        "additionalProperties": False,
        "title": "global_defaults_options",
        "description": "Set global defaults for input and output features",
    }


@DeveloperAPI
class GBMDefaultsField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(GBMDefaultsConfig)

    @staticmethod
    def _jsonschema_type_mapping():
        return get_gbm_defaults_jsonschema()



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import ACCURACY, CATEGORY, MODEL_ECD, MODEL_GBM, SOFTMAX_CROSS_ENTROPY
from ludwig.schema import utils as schema_utils
from ludwig.schema.decoders.base import BaseDecoderConfig
from ludwig.schema.decoders.utils import DecoderDataclassField
from ludwig.schema.encoders.base import BaseEncoderConfig
from ludwig.schema.encoders.utils import EncoderDataclassField
from ludwig.schema.features.base import BaseInputFeatureConfig, BaseOutputFeatureConfig
from ludwig.schema.features.loss.loss import BaseLossConfig
from ludwig.schema.features.loss.utils import LossDataclassField
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import PreprocessingDataclassField
from ludwig.schema.features.utils import (
    defaults_config_registry,
    ecd_input_config_registry,
    gbm_input_config_registry,
    input_mixin_registry,
    output_config_registry,
    output_mixin_registry,
)
from ludwig.schema.metadata import FEATURE_METADATA
from ludwig.schema.metadata.parameter_metadata import INTERNAL_ONLY
from ludwig.schema.utils import BaseMarshmallowConfig, ludwig_dataclass


@DeveloperAPI
@input_mixin_registry.register(CATEGORY)
@ludwig_dataclass
class CategoryInputFeatureConfigMixin(BaseMarshmallowConfig):
    """CategoryInputFeatureConfigMixin is a dataclass that configures the parameters used in both the category
    input feature and the category global defaults section of the Ludwig Config."""

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type=CATEGORY)


@DeveloperAPI
@ludwig_dataclass
class CategoryInputFeatureConfig(BaseInputFeatureConfig, CategoryInputFeatureConfigMixin):
    """CategoryInputFeatureConfig is a dataclass that configures the parameters used for a category input
    feature."""

    encoder: BaseEncoderConfig = None


@DeveloperAPI
@ecd_input_config_registry.register(CATEGORY)
@ludwig_dataclass
class ECDCategoryInputFeatureConfig(CategoryInputFeatureConfig):
    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_ECD,
        feature_type=CATEGORY,
        default="dense",
    )


@DeveloperAPI
@gbm_input_config_registry.register(CATEGORY)
@ludwig_dataclass
class GBMCategoryInputFeatureConfig(CategoryInputFeatureConfig):
    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_GBM,
        feature_type=CATEGORY,
        default="passthrough",
    )


@DeveloperAPI
@output_mixin_registry.register(CATEGORY)
@ludwig_dataclass
class CategoryOutputFeatureConfigMixin(BaseMarshmallowConfig):
    """CategoryOutputFeatureConfigMixin is a dataclass that configures the parameters used in both the category
    output feature and the category global defaults section of the Ludwig Config."""

    decoder: BaseDecoderConfig = DecoderDataclassField(
        feature_type=CATEGORY,
        default="classifier",
    )

    loss: BaseLossConfig = LossDataclassField(
        feature_type=CATEGORY,
        default=SOFTMAX_CROSS_ENTROPY,
    )


@DeveloperAPI
@output_config_registry.register(CATEGORY)
@ludwig_dataclass
class CategoryOutputFeatureConfig(BaseOutputFeatureConfig, CategoryOutputFeatureConfigMixin):
    """CategoryOutputFeatureConfig is a dataclass that configures the parameters used for a category output
    feature."""

    calibration: bool = schema_utils.Boolean(
        default=False,
        description="Calibrate the model's output probabilities using temperature scaling.",
        parameter_metadata=FEATURE_METADATA[CATEGORY]["calibration"],
    )

    default_validation_metric: str = schema_utils.StringOptions(
        [ACCURACY],
        default=ACCURACY,
        description="Internal only use parameter: default validation metric for category output feature.",
        parameter_metadata=INTERNAL_ONLY,
    )

    dependencies: list = schema_utils.List(
        default=[],
        description="List of input features that this feature depends on.",
        parameter_metadata=FEATURE_METADATA[CATEGORY]["dependencies"],
    )

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type="category_output")

    reduce_dependencies: str = schema_utils.ReductionOptions(
        default="sum",
        description="How to reduce the dependencies of the output feature.",
        parameter_metadata=FEATURE_METADATA[CATEGORY]["reduce_dependencies"],
    )

    reduce_input: str = schema_utils.ReductionOptions(
        default="sum",
        description="How to reduce an input that is not a vector, but a matrix or a higher order tensor, on the first "
        "dimension (second if you count the batch dimension)",
        parameter_metadata=FEATURE_METADATA[CATEGORY]["reduce_input"],
    )

    top_k: int = schema_utils.NonNegativeInteger(
        default=3,
        description="Determines the parameter k, the number of categories to consider when computing the top_k "
        "measure. It computes accuracy but considering as a match if the true category appears in the "
        "first k predicted categories ranked by decoder's confidence.",
        parameter_metadata=FEATURE_METADATA[CATEGORY]["top_k"],
    )


@DeveloperAPI
@defaults_config_registry.register(CATEGORY)
@ludwig_dataclass
class CategoryDefaultsConfig(CategoryInputFeatureConfigMixin, CategoryOutputFeatureConfigMixin):
    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_ECD,
        feature_type=CATEGORY,
        default="dense",
    )



import gc
import logging
import statistics
import time
from abc import ABC
from typing import Optional

import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import MAX_BATCH_SIZE_DATASET_FRACTION, MIN_POSSIBLE_BATCH_SIZE

logger = logging.getLogger(__name__)


@DeveloperAPI
class BatchSizeEvaluator(ABC):
    def select_best_batch_size(
        self,
        dataset_len: int,
        max_batch_size: Optional[int] = None,
        max_trials: int = 20,
    ) -> int:
        """Returns optimal batch size as measured by throughput (samples / sec)."""
        logger.info("Tuning batch size...")

        max_batch_size = max_batch_size or dataset_len

        def _is_valid_batch_size(batch_size):
            # make sure that batch size is valid (e.g. less than 20% of ds size and max_batch_size)
            is_smaller_than_training_set = batch_size <= MAX_BATCH_SIZE_DATASET_FRACTION * dataset_len
            is_under_max_batch_size = batch_size <= max_batch_size
            is_valid = is_smaller_than_training_set and is_under_max_batch_size
            if not is_valid:
                logger.info(
                    f"Batch size {batch_size} is invalid, must be less than or equal to "
                    f"{MAX_BATCH_SIZE_DATASET_FRACTION * 100}% dataset size "
                    f"({int(MAX_BATCH_SIZE_DATASET_FRACTION * dataset_len)} samples "
                    f"of {dataset_len}) and less than or equal to max batch size {max_batch_size}"
                )
            return is_valid

        # Set 2 as the minimum batch size to account for batch norm.
        batch_size = MIN_POSSIBLE_BATCH_SIZE

        best_samples_per_sec = 0
        best_batch_size = None
        count = 0
        while count < max_trials and _is_valid_batch_size(batch_size):
            logger.info(f"Exploring batch_size={batch_size}")
            gc.collect()

            try:
                samples_per_sec = self.evaluate(batch_size, total_steps=5)
                logger.info(f"Throughput at batch_size={batch_size}: {samples_per_sec:.5f} samples/s")
                if samples_per_sec < best_samples_per_sec:
                    # We assume that once the throughput starts degrading, it won't go up again
                    logger.info(f"Throughput decrease at batch_size={batch_size}")
                    break

                best_samples_per_sec = samples_per_sec
                best_batch_size = batch_size
                count += 1

                # double batch size
                batch_size *= 2
            except RuntimeError as e:
                # PyTorch only generates Runtime errors for CUDA OOM.
                gc.collect()
                if "CUDA out of memory" in str(e) or isinstance(e, torch.cuda.OutOfMemoryError):
                    logger.info(f"OOM at batch_size={batch_size}")
                else:
                    # Not a CUDA error
                    raise
                break

        # Ensure that some batch size is found.
        # `best_batch_size` can be None if the first batch size is invalid.
        if best_batch_size is None:
            logger.info("Could not tune batch size, using minimum batch size of 2")
            best_batch_size = MIN_POSSIBLE_BATCH_SIZE

        logger.info(f"Selected batch_size={best_batch_size}")
        return best_batch_size

    def evaluate(self, batch_size: int, total_steps: int = 5) -> float:
        """Evaluates throughput of the given batch size.

        Return:
            Median throughput in samples / sec.
        """
        durations = []
        for _ in range(total_steps):
            self.reset()
            start_ts = time.time()
            self.step(batch_size)
            durations.append(time.time() - start_ts)
        med_duration_s = statistics.median(durations)
        return batch_size / med_duration_s

    def reset(self):
        """Called at the beginning of each evaluation step."""
        pass

    def step(self, batch_size: int):
        """Called each step to evaluate the given batch size."""
        raise NotImplementedError("`step` must be implemented by concrete evaluator.")



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import pandas as pd

from ludwig.datasets.loaders.dataset_loader import DatasetLoader


class CreditCardFraudLoader(DatasetLoader):
    def transform_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        processed_df = super().transform_dataframe(dataframe)
        # Train/Test split like https://www.kaggle.com/competitions/1056lab-fraud-detection-in-credit-card/overview
        processed_df = processed_df.sort_values(by=["Time"])
        processed_df.loc[:198365, "split"] = 0
        processed_df.loc[198365:, "split"] = 2
        processed_df.split = processed_df.split.astype(int)
        return processed_df



import logging
from abc import ABC
from importlib import import_module
from typing import Dict

from ludwig.hyperopt.registry import register_search_algorithm

logger = logging.getLogger(__name__)


def _is_package_installed(package_name: str, search_algo_name: str) -> bool:
    try:
        import_module(package_name)
        return True
    except ImportError:
        raise ImportError(
            f"Search algorithm {search_algo_name} requires package {package_name}, however package is not installed."
            " Please refer to Ray Tune documentation for packages required for this search algorithm."
        )


class SearchAlgorithm(ABC):
    def __init__(self, search_alg_dict: Dict) -> None:
        self.search_alg_dict = search_alg_dict
        self.random_seed_attribute_name = None

    def check_for_random_seed(self, ludwig_random_seed: int) -> None:
        if self.random_seed_attribute_name not in self.search_alg_dict:
            self.search_alg_dict[self.random_seed_attribute_name] = ludwig_random_seed


@register_search_algorithm("random")
@register_search_algorithm("variant_generator")
class BasicVariantSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        super().__init__(search_alg_dict)
        self.random_seed_attribute_name = "random_state"


@register_search_algorithm("hyperopt")
class HyperoptSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("hyperopt", "hyperopt")
        super().__init__(search_alg_dict)
        self.random_seed_attribute_name = "random_state_seed"


@register_search_algorithm("bohb")
class BOHBSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("hpbandster", "bohb")
        _is_package_installed("ConfigSpace", "bohb")
        super().__init__(search_alg_dict)
        self.random_seed_attribute_name = "seed"


@register_search_algorithm("ax")
class AxSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("sqlalchemy", "ax")
        _is_package_installed("ax", "ax")
        super().__init__(search_alg_dict)

    # override parent method, this search algorithm does not support
    # setting random seed
    def check_for_random_seed(self, ludwig_random_seed: int) -> None:
        pass


@register_search_algorithm("bayesopt")
class BayesOptSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("bayes_opt", "bayesopt")
        super().__init__(search_alg_dict)
        self.random_seed_attribute_name = "random_state"


@register_search_algorithm("blendsearch")
class BlendsearchSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("flaml", "blendsearch")
        super().__init__(search_alg_dict)

    # override parent method, this search algorithm does not support
    # setting random seed
    def check_for_random_seed(self, ludwig_random_seed: int) -> None:
        pass


@register_search_algorithm("cfo")
class CFOSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("flaml", "cfo")
        super().__init__(search_alg_dict)
        self.random_seed_attribute_name = "seed"

    # override parent method, this search algorithm does not support
    # setting random seed
    def check_for_random_seed(self, ludwig_random_seed: int) -> None:
        pass


@register_search_algorithm("dragonfly")
class DragonflySA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("dragonfly", "dragonfly")
        super().__init__(search_alg_dict)
        self.random_seed_attribute_name = "random_state_seed"


@register_search_algorithm("hebo")
class HEBOSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("hebo", "hebo")
        super().__init__(search_alg_dict)
        self.random_seed_attribute_name = "random_state_seed"


@register_search_algorithm("skopt")
class SkoptSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("skopt", "skopt")
        super().__init__(search_alg_dict)

    # override parent method, this search algorithm does not support
    # setting random seed
    def check_for_random_seed(self, ludwig_random_seed: int) -> None:
        pass


@register_search_algorithm("nevergrad")
class NevergradSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("nevergrad", "nevergrad")
        super().__init__(search_alg_dict)

    # override parent method, this search algorithm does not support
    # setting random seed
    def check_for_random_seed(self, ludwig_random_seed: int) -> None:
        pass


@register_search_algorithm("optuna")
class OptunaSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("optuna", "optuna")
        super().__init__(search_alg_dict)
        self.random_seed_attribute_name = "seed"


@register_search_algorithm("zoopt")
class ZooptSA(SearchAlgorithm):
    def __init__(self, search_alg_dict: Dict) -> None:
        _is_package_installed("zoopt", "zoopt")
        super().__init__(search_alg_dict)

    # override parent method, this search algorithm does not support
    # setting random seed
    def check_for_random_seed(self, ludwig_random_seed: int) -> None:
        pass



from typing import List, Tuple, Union

import pandas as pd
import torch

try:
    import dask.dataframe as dd

    DataFrame = Union[pd.DataFrame, dd.DataFrame]
    Series = Union[pd.Series, dd.Series]
except ImportError:
    DataFrame = pd.DataFrame
    Series = pd.Series

# torchaudio.load returns the audio tensor and the sampling rate as a tuple.
TorchAudioTuple = Tuple[torch.Tensor, int]
TorchscriptPreprocessingInput = Union[List[str], List[torch.Tensor], List[TorchAudioTuple], torch.Tensor]
TorchDevice = Union[str, torch.device]



#! /usr/bin/env python
# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import re
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.utils.fs_utils import checksum
from ludwig.utils.types import DataFrame


def alphanum(v):
    """Filters a string to only its alphanumeric characters."""
    return re.sub(r"\W+", "", v)


@DeveloperAPI
class CacheableDataset(ABC):
    name: str
    checksum: str

    @abstractmethod
    def get_cache_path(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def get_cache_directory(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def unwrap(self) -> Union[str, DataFrame]:
        raise NotImplementedError()


@DeveloperAPI
@dataclass
class CacheableDataframe(CacheableDataset):
    df: DataFrame
    name: str
    checksum: str

    def get_cache_path(self) -> str:
        return alphanum(self.name)

    def get_cache_directory(self) -> str:
        return os.getcwd()

    def unwrap(self) -> Union[str, DataFrame]:
        return self.df


@DeveloperAPI
@dataclass
class CacheablePath(CacheableDataset):
    path: str

    @property
    def name(self) -> str:
        return Path(self.path).stem

    @property
    def checksum(self) -> str:
        return checksum(self.path)

    def get_cache_path(self) -> str:
        return self.name

    def get_cache_directory(self) -> str:
        return os.path.dirname(self.path)

    def unwrap(self) -> Union[str, DataFrame]:
        return self.path


CacheInput = Union[str, DataFrame, CacheableDataset]


def wrap(dataset: Optional[CacheInput]) -> CacheableDataset:
    if dataset is None:
        return None

    if isinstance(dataset, CacheableDataset):
        return dataset
    if isinstance(dataset, str):
        return CacheablePath(path=dataset)

    # TODO(travis): could try hashing the in-memory dataset, but this is tricky for Dask
    checksum = str(uuid.uuid1())
    name = checksum
    return CacheableDataframe(df=dataset, name=name, checksum=checksum)



#! /usr/bin/env python
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Dict

import numpy as np


class Batcher(ABC):
    @abstractmethod
    def next_batch(self) -> Dict[str, np.ndarray]:
        raise NotImplementedError()

    @abstractmethod
    def last_batch(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def set_epoch(self, epoch: int, batch_size: int):
        raise NotImplementedError()



#! /usr/bin/env python
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import contextlib
import logging
import os

from ludwig.api_annotations import DeveloperAPI
from ludwig.backend.base import Backend, LocalBackend
from ludwig.utils.horovod_utils import has_horovodrun

logger = logging.getLogger(__name__)


# TODO: remove LOCAL_BACKEND as a global constant, replace with singleton LocalBackend.shared_instance().
LOCAL_BACKEND = LocalBackend.shared_instance()


LOCAL = "local"
DASK = "dask"
HOROVOD = "horovod"
RAY = "ray"

ALL_BACKENDS = [LOCAL, DASK, HOROVOD, RAY]


def _has_ray():
    # Temporary workaround to prevent tests from automatically using the Ray backend. Taken from
    # https://stackoverflow.com/questions/25188119/test-if-code-is-executed-from-within-a-py-test-session
    if "PYTEST_CURRENT_TEST" in os.environ:
        return False

    try:
        import ray
    except ImportError:
        return False

    if ray.is_initialized():
        return True

    try:
        ray.init("auto", ignore_reinit_error=True)
        return True
    except Exception:
        return False


def get_local_backend(**kwargs):
    return LocalBackend(**kwargs)


def create_horovod_backend(**kwargs):
    from ludwig.backend.horovod import HorovodBackend

    return HorovodBackend(**kwargs)


def create_ray_backend(**kwargs):
    from ludwig.backend.ray import RayBackend

    return RayBackend(**kwargs)


backend_registry = {
    LOCAL: get_local_backend,
    HOROVOD: create_horovod_backend,
    RAY: create_ray_backend,
    None: get_local_backend,
}


@DeveloperAPI
def create_backend(type, **kwargs):
    if isinstance(type, Backend):
        return type

    if type is None and _has_ray():
        type = RAY
    elif type is None and has_horovodrun():
        type = HOROVOD

    return backend_registry[type](**kwargs)


@DeveloperAPI
def initialize_backend(backend):
    if isinstance(backend, dict):
        backend = create_backend(**backend)
    else:
        backend = create_backend(backend)
    backend.initialize()
    return backend


@contextlib.contextmanager
def provision_preprocessing_workers(backend):
    if backend.BACKEND_TYPE == RAY:
        with backend.provision_preprocessing_workers():
            yield
    else:
        yield



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import platform
import sys
from typing import List, Union

import numpy as np
import pandas as pd
import pytest
import torch
from packaging.version import parse as parse_version

from ludwig.api import LudwigModel
from ludwig.constants import BATCH_SIZE, NAME, PREDICTIONS, TRAINER
from ludwig.utils.neuropod_utils import export_neuropod
from tests.integration_tests.utils import (
    binary_feature,
    category_feature,
    generate_data,
    LocalTestBackend,
    number_feature,
)


@pytest.mark.skipif(platform.system() == "Windows", reason="Neuropod is not supported on Windows")
@pytest.mark.skipif(sys.version_info >= (3, 9), reason="Neuropod does not support Python 3.9")
@pytest.mark.skipif(
    parse_version(torch.__version__) >= parse_version("1.12"), reason="Neuropod does not support PyTorch >= 1.12"
)
def test_neuropod_torchscript(csv_filename, tmpdir):
    data_csv_path = os.path.join(tmpdir, csv_filename)

    # Configure features to be tested:
    bin_str_feature = binary_feature()
    input_features = [
        bin_str_feature,
        # binary_feature(),
        number_feature(),
        category_feature(encoder={"vocab_size": 3}),
        # TODO: future support
        # sequence_feature(vocab_size=3),
        # text_feature(vocab_size=3),
        # vector_feature(),
        # image_feature(image_dest_folder),
        # audio_feature(audio_dest_folder),
        # timeseries_feature(),
        # date_feature(),
        # h3_feature(),
        # set_feature(vocab_size=3),
        # bag_feature(vocab_size=3),
    ]
    output_features = [
        bin_str_feature,
        # binary_feature(),
        number_feature(),
        category_feature(decoder={"vocab_size": 3}, output_feature=True),
        # TODO: future support
        # sequence_feature(vocab_size=3),
        # text_feature(vocab_size=3),
        # set_feature(vocab_size=3),
        # vector_feature()
    ]
    backend = LocalTestBackend()
    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    # Generate training data
    training_data_csv_path = generate_data(input_features, output_features, data_csv_path)

    # Convert bool values to strings, e.g., {'Yes', 'No'}
    df = pd.read_csv(training_data_csv_path)
    false_value, true_value = "No", "Yes"
    df[bin_str_feature[NAME]] = df[bin_str_feature[NAME]].map(lambda x: true_value if x else false_value)
    df.to_csv(training_data_csv_path)

    # Train Ludwig (Pythonic) model:
    ludwig_model = LudwigModel(config, backend=backend)
    ludwig_model.train(
        dataset=training_data_csv_path,
        skip_save_training_description=True,
        skip_save_training_statistics=True,
        skip_save_model=True,
        skip_save_progress=True,
        skip_save_log=True,
        skip_save_processed_input=True,
    )

    # Obtain predictions from Python model
    preds_dict, _ = ludwig_model.predict(dataset=training_data_csv_path, return_type=dict)

    # Create graph inference model (Torchscript) from trained Ludwig model.
    neuropod_path = os.path.join(tmpdir, "neuropod")
    export_neuropod(ludwig_model, neuropod_path)

    from neuropod.loader import load_neuropod

    neuropod_module = load_neuropod(neuropod_path)

    def to_input(s: pd.Series) -> Union[List[str], torch.Tensor]:
        if s.dtype == "object":
            return np.array(s.to_list())
        return s.to_numpy().astype(np.float32)

    df = pd.read_csv(training_data_csv_path)
    inputs = {name: to_input(df[feature.column]) for name, feature in ludwig_model.model.input_features.items()}
    outputs = neuropod_module.infer(inputs)

    # Compare results from Python trained model against Neuropod
    assert len(preds_dict) == len(outputs)
    for feature_name, feature_outputs_expected in preds_dict.items():
        assert feature_name in outputs

        output_values_expected = feature_outputs_expected[PREDICTIONS]
        output_values = outputs[feature_name]
        if output_values.dtype.type in {np.string_, np.str_}:
            # Strings should match exactly
            assert np.all(output_values == output_values_expected), f"feature: {feature_name}, output: predictions"
        else:
            assert np.allclose(output_values, output_values_expected), f"feature: {feature_name}, output: predictions"



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging

import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BINARY, NUMBER, VECTOR
from ludwig.encoders.base import Encoder
from ludwig.encoders.registry import register_encoder
from ludwig.modules.fully_connected_modules import FCStack
from ludwig.schema.encoders.base import DenseEncoderConfig, PassthroughEncoderConfig

logger = logging.getLogger(__name__)


@DeveloperAPI
@register_encoder("passthrough", [BINARY, NUMBER, VECTOR])
class PassthroughEncoder(Encoder):
    def __init__(self, input_size=1, encoder_config=None, **kwargs):
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")
        self.input_size = input_size

    def forward(self, inputs, mask=None):
        """
        :param inputs: The inputs fed into the encoder.
               Shape: [batch x 1], type tf.float32
        """
        return {"encoder_output": inputs}

    @staticmethod
    def get_schema_cls():
        return PassthroughEncoderConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.input_size])

    @property
    def output_shape(self) -> torch.Size:
        return self.input_shape


@DeveloperAPI
@register_encoder("dense", [BINARY, NUMBER, VECTOR])
class DenseEncoder(Encoder):
    def __init__(
        self,
        input_size,
        layers=None,
        num_layers=1,
        output_size=256,
        use_bias=True,
        weights_initializer="xavier_uniform",
        bias_initializer="zeros",
        norm=None,
        norm_params=None,
        activation="relu",
        dropout=0,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")
        self.input_size = input_size

        logger.debug("  FCStack")
        self.fc_stack = FCStack(
            first_layer_input_size=input_size,
            layers=layers,
            num_layers=num_layers,
            default_output_size=output_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
        )

    def forward(self, inputs, training=None, mask=None):
        """
        :param inputs: The inputs fed into the encoder.
               Shape: [batch x 1], type tf.float32
        """
        return {"encoder_output": self.fc_stack(inputs)}

    @staticmethod
    def get_schema_cls():
        return DenseEncoderConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.input_size])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.fc_stack.layers[-1]["output_size"]])



import os

from ludwig.backend.ray import initialize_ray

try:
    import ray
except ImportError:
    raise ImportError(" ray is not installed. " "In order to use auto_train please run " "pip install ludwig[ray]")


def _ray_init():
    if ray.is_initialized():
        return

    # Forcibly terminate trial requested to stop after this amount of time passes
    os.environ.setdefault("TUNE_FORCE_TRIAL_CLEANUP_S", "120")

    initialize_ray()



import random
import tempfile

import numpy as np
import pytest
import torch

from ludwig.api import LudwigModel
from ludwig.constants import TRAINER
from ludwig.data.preprocessing import preprocess_for_training
from ludwig.utils.data_utils import read_csv
from ludwig.utils.torch_utils import get_torch_device
from tests.integration_tests.utils import (
    binary_feature,
    category_feature,
    date_feature,
    generate_data,
    image_feature,
    LocalTestBackend,
    number_feature,
    sequence_feature,
    set_feature,
)

DEVICE = get_torch_device()
BATCH_SIZE = 32
RANDOM_SEED = 42
IMAGE_DIR = tempfile.mkdtemp()


@pytest.mark.parametrize(
    "input_features,output_features",
    [
        (
            [number_feature(encoder={"num_layers": 2, "type": "dense"}, preprocessing={"normalization": "zscore"})],
            [number_feature()],
        ),
        ([image_feature(IMAGE_DIR, encoder={"type": "stacked_cnn"})], [number_feature()]),
        ([image_feature(IMAGE_DIR, encoder={"type": "stacked_cnn"})], [category_feature(output_feature=True)]),
        (
            [category_feature(encoder={"representation": "dense"})],
            [number_feature(decoder={"type": "regressor", "num_fc_layers": 5}, loss={"type": "mean_squared_error"})],
        ),
        ([date_feature()], [binary_feature()]),
        ([sequence_feature(encoder={"type": "parallel_cnn", "cell_type": "gru"})], [binary_feature()]),
        ([set_feature()], [set_feature(output_feature=True)]),
    ],
)
def test_regularizers(
    input_features,
    output_features,
):

    np.random.seed(RANDOM_SEED)
    torch.manual_seed(RANDOM_SEED)
    random.seed(0)

    data_file = generate_data(input_features, output_features, num_examples=BATCH_SIZE)
    data_df = read_csv(data_file)

    regularizer_losses = []
    for regularization_type in [None, "l1", "l2", "l1_l2"]:

        config = {
            "input_features": input_features,
            "output_features": output_features,
            "combiner": {"type": "concat", "output_size": 14},
            TRAINER: {
                "epochs": 2,
                "regularization_type": regularization_type,
                "regularization_lambda": 0.1,
                "batch_size": BATCH_SIZE,  # fix the batch size to ensure deterministic results
            },
        }

        backend = LocalTestBackend()
        model = LudwigModel(config, backend=backend)
        processed_data_df, _, _, _ = preprocess_for_training(model.config, data_df, backend=backend)
        with processed_data_df.initialize_batcher(batch_size=BATCH_SIZE) as batcher:
            batch = batcher.next_batch()

        _, _, _ = model.train(
            training_set=data_df,
            skip_save_processed_input=True,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
        )

        inputs = {
            i_feat.feature_name: torch.from_numpy(np.array(batch[i_feat.proc_column], copy=True)).to(DEVICE)
            for i_feat in model.model.input_features.values()
        }
        targets = {
            o_feat.feature_name: torch.from_numpy(np.array(batch[o_feat.proc_column], copy=True)).to(DEVICE)
            for o_feat in model.model.output_features.values()
        }
        predictions = model.model((inputs, targets))

        loss, _ = model.model.train_loss(targets, predictions, regularization_type, 0.1)
        regularizer_losses.append(loss)

    # Regularizer_type=None has lowest regularizer loss
    assert min(regularizer_losses) == regularizer_losses[0]

    # l1, l2 and l1_l2 should be greater than zero
    assert torch.all(torch.tensor([t - regularizer_losses[0] > 0.0 for t in regularizer_losses[1:]]))

    # using default setting l1 + l2 == l1_l2 losses
    assert torch.isclose(
        regularizer_losses[1] + regularizer_losses[2] - regularizer_losses[0], regularizer_losses[3], rtol=0.1
    )



from dataclasses import Field
from typing import Any, Dict, List, Type, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import TYPE
from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import DECODER_METADATA
from ludwig.schema.metadata.parameter_metadata import convert_metadata_to_json
from ludwig.utils.registry import Registry

decoder_config_registry = Registry()


@DeveloperAPI
def register_decoder_config(name: str, features: Union[str, List[str]]):
    if isinstance(features, str):
        features = [features]

    def wrap(cls):
        for feature in features:
            feature_registry = decoder_config_registry.get(feature, {})
            feature_registry[name] = cls
            decoder_config_registry[feature] = feature_registry
        return cls

    return wrap


@DeveloperAPI
def get_decoder_cls(feature: str, name: str):
    return decoder_config_registry[feature][name]


@DeveloperAPI
def get_decoder_classes(feature: str):
    return decoder_config_registry[feature]


@DeveloperAPI
def get_decoder_descriptions(feature_type: str):
    """This function returns a dictionary of decoder descriptions available at the type selection.

    The process works as follows - 1) Get a dictionary of valid decoders from the decoder config registry,
    but inverse the key/value pairs since we need to index `valid_decoders` later with an altered version
    of the decoder config class name. 2) Loop through Decoder Metadata entries, if a metadata entry has a
    decoder name that matches a valid decoder, add the description metadata to the output dictionary.

    Args:
        feature_type (str): The feature type to get decoder descriptions for
    Returns:
        dict: A dictionary of decoder descriptions
    """
    output = {}
    valid_decoders = {
        cls.module_name() if hasattr(cls, "module_name") else None: registered_name
        for registered_name, cls in get_decoder_classes(feature_type).items()
    }

    for k, v in DECODER_METADATA.items():
        if k in valid_decoders.keys():
            output[valid_decoders[k]] = convert_metadata_to_json(v[TYPE])

    return output


@DeveloperAPI
def get_decoder_conds(feature_type: str) -> List[Dict[str, Any]]:
    """Returns a JSON schema of conditionals to validate against decoder types for specific feature types."""
    conds = []
    for decoder in get_decoder_classes(feature_type):
        decoder_cls = get_decoder_cls(feature_type, decoder)
        other_props = schema_utils.unload_jsonschema_from_marshmallow_class(decoder_cls)["properties"]
        schema_utils.remove_duplicate_fields(other_props)
        decoder_cond = schema_utils.create_cond(
            {"type": decoder},
            other_props,
        )
        conds.append(decoder_cond)
    return conds


@DeveloperAPI
def DecoderDataclassField(feature_type: str, default: str) -> Field:
    """Custom dataclass field that when used inside a dataclass will allow the user to specify a decoder config.

    Returns: Initialized dataclass field that converts an untyped dict with params to a decoder config.
    """
    decoder_registry = get_decoder_classes(feature_type)

    class DecoderSelection(schema_utils.TypeSelection):
        def __init__(self):
            super().__init__(registry=decoder_registry, default_value=default)

        def get_schema_from_registry(self, key: str) -> Type[schema_utils.BaseMarshmallowConfig]:
            return get_decoder_cls(feature_type, key)

        @staticmethod
        def _jsonschema_type_mapping():
            return {
                "type": "object",
                "properties": {
                    "type": {
                        "type": "string",
                        "enum": list(decoder_registry.keys()),
                        "enumDescriptions": get_decoder_descriptions(feature_type),
                        "default": default,
                    },
                },
                "title": "decoder_options",
                "allOf": get_decoder_conds(feature_type),
            }

    return DecoderSelection().get_default_field()



from datetime import datetime
from typing import Dict, Optional

import pandas as pd
import torch

from ludwig.constants import (
    AUDIO,
    BAG,
    BINARY,
    CATEGORY,
    COLUMN,
    DATE,
    IMAGE,
    NAME,
    POSTPROCESSOR,
    PREDICTOR,
    PREPROCESSOR,
    SEQUENCE,
    SET,
    TEXT,
    TIMESERIES,
    TYPE,
    VECTOR,
)
from ludwig.types import FeatureConfigDict, ModelConfigDict
from ludwig.utils.audio_utils import read_audio_from_path
from ludwig.utils.date_utils import create_vector_from_datetime_obj
from ludwig.utils.image_utils import read_image_from_path
from ludwig.utils.torch_utils import place_on_device
from ludwig.utils.types import TorchDevice, TorchscriptPreprocessingInput

FEATURES_TO_CAST_AS_STRINGS = {BINARY, CATEGORY, BAG, SET, TEXT, SEQUENCE, TIMESERIES, VECTOR}


def get_filename_from_stage(stage: str, device: TorchDevice) -> str:
    """Returns the filename for a stage of inference."""
    if stage not in [PREPROCESSOR, PREDICTOR, POSTPROCESSOR]:
        raise ValueError(f"Invalid stage: {stage}.")
    # device is only tracked for predictor stage
    if stage == PREDICTOR:
        return f"inference_{stage}-{device}.pt"
    else:
        return f"inference_{stage}.pt"


def to_inference_module_input_from_dataframe(
    dataset: pd.DataFrame, config: ModelConfigDict, load_paths: bool = False, device: Optional[torch.device] = None
) -> Dict[str, TorchscriptPreprocessingInput]:
    """Converts a pandas DataFrame to be compatible with a torchscripted InferenceModule forward pass."""
    inputs = {}
    for if_config in config["input_features"]:
        feature_inputs = to_inference_model_input_from_series(
            dataset[if_config[COLUMN]],
            if_config[TYPE],
            load_paths=load_paths,
            feature_config=if_config,
        )
        feature_inputs = place_on_device(feature_inputs, device)
        inputs[if_config[NAME]] = feature_inputs
    return inputs


def to_inference_model_input_from_series(
    s: pd.Series, feature_type: str, load_paths: bool = False, feature_config: Optional[FeatureConfigDict] = None
) -> TorchscriptPreprocessingInput:
    """Converts a pandas Series to be compatible with a torchscripted InferenceModule forward pass."""
    if feature_type == IMAGE:
        if load_paths:
            return [read_image_from_path(v) if isinstance(v, str) else v for v in s]
    elif feature_type == AUDIO:
        if load_paths:
            return [read_audio_from_path(v) if isinstance(v, str) else v for v in s]
    elif feature_type == DATE:
        if feature_config is None:
            raise ValueError('"date" feature type requires the associated feature config to be provided.')
        datetime_format = feature_config["preprocessing"]["datetime_format"]
        return [torch.tensor(create_vector_from_datetime_obj(datetime.strptime(v, datetime_format))) for v in s]
    elif feature_type in FEATURES_TO_CAST_AS_STRINGS:
        return s.astype(str).to_list()
    return torch.from_numpy(s.to_numpy())



from typing import Any, Dict, List, Type

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import TYPE
from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import COMBINER_METADATA
from ludwig.schema.metadata.parameter_metadata import convert_metadata_to_json, ParameterMetadata
from ludwig.utils.registry import Registry

DEFAULT_VALUE = "concat"
DESCRIPTION = "Select the combiner type."

combiner_registry = Registry()


@DeveloperAPI
def register_combiner(name: str):
    def wrap(cls):
        combiner_registry[name] = cls
        return cls

    return wrap


@DeveloperAPI
def get_combiner_registry():
    return combiner_registry


@DeveloperAPI
def get_combiner_jsonschema():
    """Returns a JSON schema structured to only require a `type` key and then conditionally apply a corresponding
    combiner's field constraints."""
    combiner_types = sorted(list(combiner_registry.keys()))
    parameter_metadata = convert_metadata_to_json(
        ParameterMetadata.from_dict(
            {
                "commonly_used": True,
                "expected_impact": 3,
                "ui_display_name": "Combiner Type",
            }
        )
    )
    return {
        "type": "object",
        "properties": {
            "type": {
                "type": "string",
                "enum": combiner_types,
                "enumDescriptions": get_combiner_descriptions(),
                "default": DEFAULT_VALUE,
                "title": "combiner_options",
                "description": DESCRIPTION,
                "parameter_metadata": parameter_metadata,
            },
        },
        "allOf": get_combiner_conds(),
        "required": ["type"],
    }


@DeveloperAPI
def get_combiner_descriptions():
    """This function returns a dictionary of combiner descriptions available at the type selection.

    The process works as follows - 1) Get a dictionary of valid combiners from the combiner config registry,
    but inverse the key/value pairs since we need to index `valid_combiners` later with an altered version
    of the combiner config class name. 2) Loop through Combiner Metadata entries, if a metadata entry has a
    combiner name that matches a valid combiner, add the description metadata to the output dictionary.

    Returns:
        dict: A dictionary of combiner descriptions.
    """
    output = {}
    combiners = {cls.__name__: registered_name for registered_name, cls in combiner_registry.items()}

    for k, v in COMBINER_METADATA.items():
        if k in combiners.keys():
            output[combiners[k]] = convert_metadata_to_json(v[TYPE])

    return output


@DeveloperAPI
def get_combiner_conds() -> List[Dict[str, Any]]:
    """Returns a list of if-then JSON clauses for each combiner type in `combiner_registry` and its properties'
    constraints."""
    combiner_types = sorted(list(combiner_registry.keys()))
    conds = []
    for combiner_type in combiner_types:
        combiner_cls = combiner_registry[combiner_type]
        schema_cls = combiner_cls.get_schema_cls()
        combiner_schema = schema_utils.unload_jsonschema_from_marshmallow_class(schema_cls)
        combiner_props = combiner_schema["properties"]
        schema_utils.remove_duplicate_fields(combiner_props)
        combiner_cond = schema_utils.create_cond({"type": combiner_type}, combiner_props)
        conds.append(combiner_cond)
    return conds


class CombinerSelection(schema_utils.TypeSelection):
    def __init__(self):
        # For registration of all combiners
        import ludwig.combiners.combiners  # noqa

        super().__init__(registry=combiner_registry, default_value=DEFAULT_VALUE, description=DESCRIPTION)

    def get_schema_from_registry(self, key: str) -> Type[schema_utils.BaseMarshmallowConfig]:
        return self.registry[key].get_schema_cls()

    @staticmethod
    def _jsonschema_type_mapping():
        return get_combiner_jsonschema()



import torch
import torch.nn as nn

from ludwig.modules.initializer_modules import get_initializer
from ludwig.utils.torch_utils import get_torch_device

DEVICE = "cuda:0" if get_torch_device() == "cuda" else "cpu"


def test_get_initializer():
    """Currently only checks for when the parameters are default case."""
    tensor_size = (2, 3)

    # Test for when the parameters are default
    torch.random.manual_seed(0)
    initialized_tensor = get_initializer("xavier_uniform")(*tensor_size, device=DEVICE)

    # Check that the tensor using the expected initialization and the same seed is identical
    default_initializer = nn.init.xavier_uniform_
    torch.random.manual_seed(0)
    default_tensor = default_initializer(torch.empty(*tensor_size, device=DEVICE))
    assert torch.equal(initialized_tensor, default_tensor)



# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
from typing import Any, List, Optional

import torch

try:
    import horovod.torch

    _HVD = horovod.torch
except (ModuleNotFoundError, ImportError):
    _HVD = None


def initialize_horovod():
    if not _HVD:
        raise ValueError(
            "Horovod backend specified, "
            "but cannot import `horovod.torch`. "
            "Install Horovod following the instructions at: "
            "https://github.com/horovod/horovod"
        )
    _HVD.init()
    return _HVD


def has_horovodrun():
    """Returns True if running with `horovodrun` using Gloo or OpenMPI."""
    return "OMPI_COMM_WORLD_RANK" in os.environ or "HOROVOD_RANK" in os.environ


def gather_all_tensors(result: torch.Tensor, group: Optional[Any] = None) -> List[torch.Tensor]:
    """Function to gather all tensors from several processes onto a list that is broadcast to all processes.

    Works on tensors that have the same number of dimensions, but where each dimension may differ. In this case
    tensors are padded, gathered and then trimmed to secure equal workload for all processes.

    :param result: the value to sync
    :param group: the process group to gather results from (not supported: always uses world)

    :return: list with size equal to the process group where gathered_result[i]
             corresponds to result tensor from process i
    """
    if group is not None:
        raise ValueError("Horovod does not support allgather using a subcommunicator at this time. " "Unset `group`.")

    if _HVD is None or not _HVD.is_initialized():
        return [result]

    if len(result.shape) == 0:
        # Convert scalars to single dimension tensors
        result = result.reshape(1)

    is_bool = False
    if result.dtype == torch.bool:
        # need to convert to int due to Horovod limitation
        result = result.int()
        is_bool = True

    # Add extra dimension to the tensors to be gathered
    result = result.unsqueeze(0)

    # sync and gather all
    gathered = _HVD.allgather(result)
    gathered_result = list(gathered.split(1, dim=0))

    if is_bool:
        # convert back if needed
        gathered_result = [t.bool() for t in gathered_result]

    return gathered_result


def is_distributed_available() -> bool:
    return _HVD is not None and (_HVD.is_initialized() or os.environ.get("HOROVOD_RANK"))



from typing import Union

import pytest
import torch

from ludwig.encoders.image.base import MLPMixerEncoder, ResNetEncoder, Stacked2DCNN, ViTEncoder
from ludwig.encoders.image.torchvision import (
    TVAlexNetEncoder,
    TVConvNeXtEncoder,
    TVDenseNetEncoder,
    TVEfficientNetEncoder,
    TVGoogLeNetEncoder,
    TVInceptionV3Encoder,
    TVMaxVitEncoder,
    TVMNASNetEncoder,
    TVMobileNetV2Encoder,
    TVMobileNetV3Encoder,
    TVRegNetEncoder,
    TVResNetEncoder,
    TVResNeXtEncoder,
    TVShuffleNetV2Encoder,
    TVSqueezeNetEncoder,
    TVSwinTransformerEncoder,
    TVVGGEncoder,
    TVViTEncoder,
    TVWideResNetEncoder,
)
from ludwig.utils.image_utils import torchvision_model_registry
from ludwig.utils.misc_utils import set_random_seed
from tests.integration_tests.parameter_update_utils import check_module_parameters_updated

RANDOM_SEED = 1919


@pytest.mark.parametrize("height,width,num_conv_layers,num_channels", [(224, 224, 5, 3)])
def test_stacked2d_cnn(height: int, width: int, num_conv_layers: int, num_channels: int):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    stacked_2d_cnn = Stacked2DCNN(
        height=height, width=width, num_conv_layers=num_conv_layers, num_channels=num_channels
    )
    inputs = torch.rand(2, num_channels, height, width)
    outputs = stacked_2d_cnn(inputs)
    assert outputs["encoder_output"].shape[1:] == stacked_2d_cnn.output_shape

    # check for parameter updating
    target = torch.randn(outputs["encoder_output"].shape)
    fpc, tpc, upc, not_updated = check_module_parameters_updated(stacked_2d_cnn, (inputs,), target)

    assert tpc == upc, f"Not all expected parameters updated.  Parameters not updated {not_updated}."


@pytest.mark.parametrize("height,width,num_channels", [(224, 224, 1), (224, 224, 3)])
def test_resnet_encoder(height: int, width: int, num_channels: int):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    resnet = ResNetEncoder(height=height, width=width, num_channels=num_channels)
    inputs = torch.rand(2, num_channels, height, width)
    outputs = resnet(inputs)
    assert outputs["encoder_output"].shape[1:] == resnet.output_shape

    # check for parameter updating
    target = torch.randn(outputs["encoder_output"].shape)
    fpc, tpc, upc, not_updated = check_module_parameters_updated(resnet, (inputs,), target)

    assert tpc == upc, f"Not all expected parameters updated.  Parameters not updated {not_updated}."


@pytest.mark.parametrize("height,width,num_channels", [(224, 224, 3)])
def test_mlp_mixer_encoder(height: int, width: int, num_channels: int):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    mlp_mixer = MLPMixerEncoder(height=height, width=width, num_channels=num_channels)
    inputs = torch.rand(2, num_channels, height, width)
    outputs = mlp_mixer(inputs)
    assert outputs["encoder_output"].shape[1:] == mlp_mixer.output_shape

    # check for parameter updating
    target = torch.randn(outputs["encoder_output"].shape)
    fpc, tpc, upc, not_updated = check_module_parameters_updated(mlp_mixer, (inputs,), target)

    assert tpc == upc, f"Not all expected parameters updated.  Parameters not updated {not_updated}."


@pytest.mark.parametrize("image_size,num_channels", [(224, 3)])
@pytest.mark.parametrize("use_pretrained", [True, False])
def test_vit_encoder(image_size: int, num_channels: int, use_pretrained: bool):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    vit = ViTEncoder(
        height=image_size,
        width=image_size,
        num_channels=num_channels,
        use_pretrained=use_pretrained,
        output_attentions=True,
    )
    inputs = torch.rand(2, num_channels, image_size, image_size)
    outputs = vit(inputs)
    assert outputs["encoder_output"].shape[1:] == vit.output_shape
    config = vit.transformer.module.config
    num_patches = (224 // config.patch_size) ** 2 + 1  # patches of the image + cls_token
    attentions = outputs["attentions"]
    assert len(attentions) == config.num_hidden_layers
    assert attentions[0].shape == torch.Size([2, config.num_attention_heads, num_patches, num_patches])

    # check for parameter updating
    target = torch.randn(outputs["encoder_output"].shape)
    fpc, tpc, upc, not_updated = check_module_parameters_updated(vit, (inputs,), target)

    assert tpc == upc, f"Not all expected parameters updated.  Parameters not updated {not_updated}."


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["alexnet"].values()])
def test_tv_alexnet_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVAlexNetEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["convnext"].values()])
def test_tv_convnext_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVConvNeXtEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["densenet"].values()])
def test_tv_densenet_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVDenseNetEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


# test only model variants that do not require large amount of memory
LOW_MEMORY_EFFICIENTNET_VARIANTS = set(torchvision_model_registry["efficientnet"].keys()) - {"b6", "b7"}


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", LOW_MEMORY_EFFICIENTNET_VARIANTS)
def test_tv_efficientnet_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVEfficientNetEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["googlenet"].values()])
def test_tv_googlenet_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVGoogLeNetEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["inceptionv3"].values()])
def test_tv_inceptionv3_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVInceptionV3Encoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["maxvit"].values()])
def test_tv_maxvit_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVMaxVitEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["mnasnet"].values()])
def test_tv_mnasnet_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVMNASNetEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["mobilenetv2"].values()])
def test_tv_mobilenetv2_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVMobileNetV2Encoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["mobilenetv3"].values()])
def test_tv_mobilenetv3_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVMobileNetV3Encoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


# test only model variants that do not require large amount of memory
LOW_MEMORY_REGNET_VARIANTS = set(torchvision_model_registry["regnet"].keys()) - {"y_128gf"}


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", LOW_MEMORY_REGNET_VARIANTS)
def test_tv_regnet_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVRegNetEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["resnet"].values()])
def test_tv_resnet_torch_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVResNetEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["resnext"].values()])
def test_tv_resnext_encoder(
    model_variant: int,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVResNeXtEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["shufflenet_v2"].values()])
def test_tv_shufflenet_v2_encoder(
    model_variant: str,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVShuffleNetV2Encoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["squeezenet"].values()])
def test_tv_squeezenet_encoder(
    model_variant: str,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVSqueezeNetEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize(
    "model_variant", [v.variant_id for v in torchvision_model_registry["swin_transformer"].values()]
)
def test_tv_swin_transformer_encoder(
    model_variant: str,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVSwinTransformerEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["vgg"].values()])
def test_tv_vgg_encoder(
    model_variant: Union[int, str],
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVVGGEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


# test only VIT model variants that do not require large amount of memory
LOW_MEMORY_VIT_VARIANTS = set(torchvision_model_registry["vit"].keys()) - {"h_14"}


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", LOW_MEMORY_VIT_VARIANTS)
def test_tv_vit_encoder(
    model_variant: str,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVViTEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape


@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("saved_weights_in_checkpoint", [True, False])
@pytest.mark.parametrize(
    "use_pretrained",
    [
        False,
    ],
)
@pytest.mark.parametrize("model_variant", [v.variant_id for v in torchvision_model_registry["wide_resnet"].values()])
def test_tv_wide_resnet_encoder(
    model_variant: str,
    use_pretrained: bool,
    saved_weights_in_checkpoint: bool,
    trainable: bool,
):
    # make repeatable
    set_random_seed(RANDOM_SEED)

    pretrained_model = TVWideResNetEncoder(
        model_variant=model_variant,
        use_pretrained=use_pretrained,
        saved_weights_in_checkpoint=saved_weights_in_checkpoint,
        trainable=trainable,
    )
    inputs = torch.rand(2, *pretrained_model.input_shape)
    outputs = pretrained_model(inputs)
    assert outputs["encoder_output"].shape[1:] == pretrained_model.output_shape



from dataclasses import Field
from typing import Type

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import utils as schema_utils
from ludwig.schema.features.loss import get_loss_classes, get_loss_cls


@DeveloperAPI
def get_loss_conds(feature_type: str):
    """Returns a JSON schema of conditionals to validate against loss types for specific feature types."""
    conds = []
    for loss in get_loss_classes(feature_type):
        loss_cls = get_loss_cls(feature_type, loss)
        other_props = schema_utils.unload_jsonschema_from_marshmallow_class(loss_cls)["properties"]
        schema_utils.remove_duplicate_fields(other_props)
        loss_cond = schema_utils.create_cond(
            {"type": loss},
            other_props,
        )
        conds.append(loss_cond)
    return conds


@DeveloperAPI
def LossDataclassField(feature_type: str, default: str) -> Field:
    loss_registry = get_loss_classes(feature_type)

    class LossSelection(schema_utils.TypeSelection):
        def __init__(self):
            super().__init__(registry=loss_registry, default_value=default)

        def get_schema_from_registry(self, key: str) -> Type[schema_utils.BaseMarshmallowConfig]:
            return get_loss_cls(feature_type, key)

        @staticmethod
        def _jsonschema_type_mapping():
            return {
                "type": "object",
                "properties": {
                    "type": {"type": "string", "enum": list(loss_registry.keys()), "default": default},
                },
                "title": "loss_options",
                "allOf": get_loss_conds(feature_type),
            }

    return LossSelection().get_default_field()
